"""Tests for the e3.testsuite.driver.diff module."""

import os.path
import shutil
import sys
import tempfile

from e3.testsuite import Testsuite as Suite
import e3.testsuite.driver.adacore as adacore
import e3.testsuite.driver.diff as diff
from e3.testsuite.result import TestStatus as Status
from e3.testsuite.testcase_finder import ParsedTest

from .utils import create_testsuite, extract_results, run_testsuite


class DiffScriptDriver(diff.DiffTestDriver):
    """Driver to check test output with DiffTestDriver."""

    helper_script = os.path.abspath(
        os.path.join(os.path.dirname(__file__), "classic-tests", "script.py")
    )

    @property
    def refine_baseline(self):
        return self.test_env.get("refine_baseline", False)

    @property
    def output_refiners(self):
        result = super().output_refiners

        for path, replacement in self.test_env.get("path_substitutions", []):
            result += [diff.ReplacePath(self.working_dir(path), replacement)]

        if self.refine_baseline:
            result += [
                diff.Substitute("baseline_to_refine", "refined"),
                diff.Substitute("actual_to_refine", "refined"),
            ]

        return result

    def set_up(self):
        super(DiffScriptDriver, self).set_up()
        self.process_args = self.test_env["process_args"]

    def run(self):
        self.shell([sys.executable, self.helper_script] + self.process_args)


def test_diff():
    """Check that DiffTestDriver works as expected."""

    class Mysuite(Suite):
        tests_subdir = "diff-tests"
        test_driver_map = {"diff-script-driver": DiffScriptDriver}

    suite = run_testsuite(Mysuite, args=["-E"], expect_failure=True)
    assert extract_results(suite) == {
        "plain-pass": Status.PASS,
        "plain-fail": Status.FAIL,
        "binary": Status.PASS,
        # Check that binary diffs are based on equally escaped outputs: the
        # b"\\xe9" process output is different from the b"\xe9" baseline.
        "binary-2": Status.FAIL,
        "regexp-pass": Status.PASS,
        "regexp-fail": Status.FAIL,
        "regexp-binary-pass": Status.PASS,
        "regexp-binary-fail": Status.FAIL,
        "missing-baseline": Status.ERROR,
        "path-substitution": Status.PASS,
        "line-endings": Status.PASS,
        "line-endings-binary": Status.PASS,
        "line-endings-strict": Status.FAIL,
        "refine-baseline": Status.PASS,
    }


def test_diff_context_size():
    """Check handling for the diff_context_size setting."""
    baseline = "".join(f"{i}\n" for i in range(100))
    actual = "".join(f"{i}\n" for i in range(100) if i != 20)

    def create_test(name, driver, test_env):
        return ParsedTest(name, driver, test_env, ".", None)

    tests = []

    # Test default behavior and overriding it with the test environment

    class MyDriver(diff.DiffTestDriver):
        def run(self):
            pass

        def compute_failures(self):
            return self.compute_diff("output", baseline, actual)

    tests += [
        create_test("default", MyDriver, {}),
        create_test("test_env", MyDriver, {"diff_context_size": 2}),
    ]

    # Test overriding through the diff_context_size property: it should
    # short-circuit the test environment.

    class MyDriver2(MyDriver):
        @property
        def diff_context_size(self):
            return 2

    tests += [
        create_test("property", MyDriver2, {}),
        create_test("property_test_env", MyDriver2, {"diff_context_size": 5}),
    ]

    # Test direct calls to DiffTestDriver.compute_diff: non-None context_size
    # argument should override the rest.

    class MyDriver3(MyDriver):
        @property
        def diff_context_size(self):
            return 2

        def compute_failures(self):
            return self.compute_diff(
                "output", baseline, actual, context_size=5
            )

    tests += [
        create_test("method", MyDriver3, {}),
        create_test("method_test_env", MyDriver3, {"diff_context_size": 10}),
    ]

    class MySuite(Suite):
        def get_test_list(self, sublist):
            return tests

    suite = run_testsuite(MySuite, args=["-j1"], expect_failure=True)
    assert extract_results(suite) == {
        "default": Status.FAIL,
        "test_env": Status.FAIL,
        "property": Status.FAIL,
        "property_test_env": Status.FAIL,
        "method": Status.FAIL,
        "method_test_env": Status.FAIL,
    }

    def check_diff(test_name, expected):
        assert suite.report_index.entries[test_name].load().diff == expected

    diff_context_1 = (
        "Diff failure: unexpected output\n"
        "--- expected\n"
        "+++ output\n"
        "@@ -20,3 +20,2 @@\n"
        " 19\n"
        "-20\n"
        " 21\n"
    )
    diff_context_2 = (
        "Diff failure: unexpected output\n"
        "--- expected\n"
        "+++ output\n"
        "@@ -19,5 +19,4 @@\n"
        " 18\n"
        " 19\n"
        "-20\n"
        " 21\n"
        " 22\n"
    )
    diff_context_5 = (
        "Diff failure: unexpected output\n"
        "--- expected\n"
        "+++ output\n"
        "@@ -16,11 +16,10 @@\n"
        " 15\n"
        " 16\n"
        " 17\n"
        " 18\n"
        " 19\n"
        "-20\n"
        " 21\n"
        " 22\n"
        " 23\n"
        " 24\n"
        " 25\n"
    )

    check_diff("default", diff_context_1)
    check_diff("test_env", diff_context_2)
    check_diff("property", diff_context_2)
    check_diff("property_test_env", diff_context_2)
    check_diff("method", diff_context_5)
    check_diff("method_test_env", diff_context_5)


def test_regexp_fullmatch():
    """Check that DiffTestDriver does a full match over regexp baselines."""
    # The "abcd" regexp baseline should not be able to match "abcde" (it used
    # to).

    class MyDriver(diff.DiffTestDriver):
        @property
        def baseline(self):
            return (None, "abcd", True)

        def run(self):
            self.output += "abcde"

    suite = run_testsuite(
        create_testsuite(["mytest"], MyDriver), expect_failure=True
    )
    assert extract_results(suite) == {"mytest": Status.FAIL}


def test_diff_rewriting():
    """Check that DiffTestDriver's rewriting feature works as expected."""
    # This testcase involves the rewriting of testcase files, so work on a
    # temporary copy.
    with tempfile.TemporaryDirectory(prefix="test_diff_rewriting") as temp_dir:
        tests_source = os.path.join(
            os.path.dirname(__file__), "diff-rewriting-tests"
        )
        tests_copy = os.path.join(temp_dir, "tests")
        shutil.copytree(tests_source, tests_copy)

        class Mysuite(Suite):
            tests_subdir = tests_copy
            test_driver_map = {
                "diff-script-driver": DiffScriptDriver,
                "adacore-driver": adacore.AdaCoreLegacyTestDriver,
            }

            def add_options(self, parser):
                parser.add_argument("--rewrite", "-r", action="store_true")

            def set_up(self):
                super(Mysuite, self).set_up()
                self.env.rewrite_baselines = self.main.args.rewrite
                self.env.discs = []
                self.env.test_environ = dict(os.environ)

        def check_test_out(test, expected_lines, encoding="utf-8"):
            filename = os.path.join(tests_copy, test, "test.out")
            if expected_lines is None:
                assert not os.path.isfile(filename)
            else:
                with open(filename, encoding=encoding) as f:
                    lines = [line.rstrip() for line in f]
                assert lines == expected_lines

        # Make sure we have the expected baselines before running the testsuite
        check_test_out("adacore", ["legacy"])
        check_test_out("plain", ["hello", "world"])
        check_test_out("regexp", ["h.l+o", "world"])
        check_test_out("xfail", ["hello", "world"])
        check_test_out("iso-8859-1", ["héllo"], encoding="utf-8")
        check_test_out("bad-utf-8", ["héllo"], encoding="utf-8")
        check_test_out("missing-baseline", None)

        # Run the testsuite in rewrite mode
        suite = run_testsuite(Mysuite, args=["-rE"], expect_failure=True)
        assert extract_results(suite) == {
            "adacore": Status.FAIL,
            "plain": Status.FAIL,
            "regexp": Status.FAIL,
            "xfail": Status.XFAIL,
            "iso-8859-1": Status.FAIL,
            "bad-utf-8": Status.ERROR,
            "missing-baseline": Status.FAIL,
        }

        # Check that non-regexp baselines were updated, except when a failure
        # is expected.
        check_test_out("adacore", ["adacore", "legacy", "driver"])
        check_test_out("plain", ["helloo", "world", "!"])
        check_test_out("regexp", ["h.l+o", "world"])
        check_test_out("xfail", ["hello", "world"])
        check_test_out("iso-8859-1", ["héllo"], encoding="iso-8859-1")
        check_test_out("bad-utf-8", ["héllo"], encoding="utf-8")
        check_test_out("missing-baseline", ["helloo", "world", "!"])


class TestDoubleDiff:
    """Check proper result constructions for multiple diff failures."""

    class MyDriver(diff.DiffTestDriver):
        def run(self):
            pass

        def compute_failures(self):
            d1 = self.compute_diff(
                None, "a\nb\nc\n", "a\nc\n", failure_message="first diff"
            )
            d2 = self.compute_diff(
                None, "1\n3\n", "1\n2\n3\n", failure_message="second diff"
            )
            return d1 + d2

    class Mysuite(Suite):
        tests_subdir = "simple-tests"

        @property
        def test_driver_map(self):
            return {"default": TestDoubleDiff.MyDriver}

    def test(self):
        suite = run_testsuite(
            self.Mysuite, args=["test1"], expect_failure=True
        )
        assert extract_results(suite) == {"test1": Status.FAIL}

        # When multiple diff failures are involved, we expect .expected/.out to
        # be empty, as this formalism assumes that a single output comparison.
        # We expect .diff to contain both diff's though.
        result = suite.report_index.entries["test1"].load()
        assert result.expected is None
        assert result.out is None
        assert result.diff is not None
        assert "first diff" in result.diff
        assert "second diff" in result.diff


def test_failure_reason():
    """Check that DiffTestDriver properly sets the DIFF failure reason."""

    class Mysuite(Suite):
        tests_subdir = "diff-tests"
        test_driver_map = {"diff-script-driver": DiffScriptDriver}

    suite = run_testsuite(
        Mysuite,
        args=["plain-pass", "plain-fail", "--gaia-output"],
        expect_failure=True,
    )
    assert extract_results(suite) == {
        "plain-pass": Status.PASS,
        "plain-fail": Status.FAIL,
    }

    with open(os.path.join("out", "new", "results")) as f:
        results = sorted(line.strip().split(":", 2) for line in f)

    assert results == [
        ["plain-fail", "DIFF", "unexpected output"],
        ["plain-pass", "OK", ""],
    ]


def test_line_by_line():
    output = "first line\nssecond line\n"
    subst = diff.PatternSubstitute("[^ ]*second", "some-content")
    lbl_subst = diff.LineByLine(subst)

    assert subst.refine(output) == "first some-content line\n"
    assert lbl_subst.refine(output) == "first line\nsome-content line\n"

    output = b"first line\nssecond line\n"
    subst = diff.PatternSubstitute(b"[^ ]*second", b"some-content")
    lbl_subst = diff.LineByLine(subst)

    assert subst.refine(output) == b"first some-content line\n"
    assert lbl_subst.refine(output) == b"first line\nsome-content line\n"


def test_pattern_substitute_callback():
    """Check that PatternSubstute accept replacement callbacks."""

    def repl(m):
        return m.group(1)

    refiner = diff.PatternSubstitute(r"[a-z]*\((.*)\)", repl)

    assert refiner.refine("foo(1, 2)") == "1, 2"



"""test.opt files processing.

This package provides a single class called OptFileParse which process the
test.opt files. These files are used mainly to tweak tests depending on the
context.
"""

from __future__ import annotations

import argparse
import logging
import os.path
import re
import sys
from typing import Dict, List, Optional, TYPE_CHECKING, Tuple, Union


# Import Literal only if type checking, as Literal was introduced in Python
# 3.8, and we need to support Python 3.7 at runtime.
if TYPE_CHECKING:  # no cover
    from typing import Literal


# Regexp that matches valid lines in test.opt files
OPTLINE_REGEXPS = re.compile(
    # Comma-separated list of discriminants
    r"^([a-zA-Z0-9!+,._-]+)"
    # The rest is optional...
    "("
    # Command
    r"\s+([a-zA-Z0-9_-]+)"
    # Optional command argument
    r"(\s+(.*))?"
    ")?$"
)


class BadFormattingError(Exception):
    """Raised when an input line is not correctly formatted."""

    pass


TAGS: Literal[0] = 0
ARG: Literal[1] = 1
OVERIDABLE: Literal[2] = 2

logger = logging.getLogger("optfileparser")


class OptFileParse:
    """test.opt parser.

    ATTRIBUTES
      system_tags: the list of tags applied to the test.opt
      is_dead: True if the test should be considered DEAD, False otherwise
    """

    def __init__(
        self,
        system_tags: Union[str, List[str]],
        filename: Union[str, List[str]],
    ) -> None:
        """Parse a test.opt file.

        :param system_tags: Either a list of tags or a string containing the
            list of tags separated by commas.
        :param filename: The test.opt to be parsed. If this is a string then
            the argument is a filename otherwise if this is a list we consider
            it is the content of the .opt file.
        """
        if isinstance(system_tags, str):
            self.system_tags = system_tags.lower().split(",")
        else:
            self.system_tags = []
            for tag in system_tags:
                self.system_tags.append(tag.lower())
        # Append 'all' to system tags
        if "all" not in system_tags:
            self.system_tags.append("all")

        self.is_dead = False
        self.__note: Optional[Tuple[List[str], str, bool]] = None
        self.__enable_note = False
        self.__matches: Dict[str, Tuple[List[str], str, bool]] = {}
        self.__parse_file(filename)

    def get_value(
        self, cmd: str, default_value: Optional[str] = None
    ) -> Optional[str]:
        """Query on the parsing result.

        :param cmd: The command on which we do the query ex: dead, cmd, out...
        :param default_value: Value returned by default.

        :return: A string or default value (None by default).

        by default the query will return ``default_value`` if there is no
        entry for the selected command.
        """
        cmd = cmd.lower()

        if self.is_dead:
            if cmd == "dead":
                return self.__matches[cmd][ARG]
            else:
                return default_value

        if cmd in self.__matches and cmd != "dead":
            return self.__matches[cmd][ARG]
        else:
            return default_value

    def get_values(
        self, default_values: Dict[str, Optional[str]]
    ) -> Dict[str, Optional[str]]:
        """Query on the parsing result.

        :param default_values: A dictionary for which keys are the commands on
            which we do the query and the associated default values.

        :return: A dictionary containing the resulting value for each command.

        Doing ``get_values({"CMD": "test.cmd", "OUT": "test.out"})`` is
        equivalent to do::

            get_value("CMD", "test.cmd")
            get_value("OUT", "test.out")
        """
        result: Dict[str, Optional[str]] = {}
        for key in default_values:
            result[key] = self.get_value(key, default_values[key])
        return result

    def get_note(self, sep: Optional[str] = None) -> Union[str, List[str]]:
        """Get the note.

        :param sep: string used to join the activating tags. Default is ",".
            If "" is specified then a list is returned.

        :return: a string (list of tags responsible for the activation of the
            test) is sep is not "" or a list.

        If there is no note then "" or [] is returned depending on the sep
        value
        """
        if sep is None:
            sep = ","

        if len(sep) == 0:
            if self.__note is not None and not self.is_dead:
                return self.__note[TAGS]
            else:
                return []
        else:
            if self.__note is not None and not self.is_dead:
                return ",".join(self.__note[TAGS])
            else:
                return ""

    # INTERNAL FUNCTIONS
    def __process_opt_line(self, line: str, lineno: int) -> None:
        """process one line of a test.opt type file.

        :raise BadFormattingError: in case the line cannot be parsed

        The format of each line is the following::

            tag1,tag2,tag3,...,tagN [COMMAND [PARAMETERS]]

        if no COMMAND is given then we assume that the command is 'DEAD false'
        """
        # Remove comments (Ada style) and trailing white characters
        processed_line = re.sub("--.*$", "", line).rstrip()

        # Line contains only comments and/or white characters so ignore it
        if not processed_line:
            return

        m = OPTLINE_REGEXPS.match(processed_line)
        if m is None:
            raise BadFormattingError(
                f"Can not parse line {lineno}: {line.rstrip()}"
            )

        # find command, tags and argument
        tags = m.group(1).split(",")

        if m.group(3):
            # Check for command
            cmd = m.group(3).lower()
            if m.group(4):
                # Get optional argument
                arg = m.group(5)
            else:
                arg = ""
        else:
            # If no command is set then the implicit command is: dead="false"
            cmd = "dead"
            arg = "false"

        if arg == "" and cmd == "dead":
            arg = "true"

        # Enable note only if a dead all is encountered
        if arg != "false" and cmd == "dead" and self.__is_all(tags):
            self.__enable_note = True

        if cmd != "required" and self.__match(tags):
            logger.debug("match: " + cmd + ", tags=" + "%s" % tags)
            if self.__is_overidable(cmd):
                self.__matches[cmd] = (tags, arg, self.__is_all(tags))

                if not self.__is_dead_cmd(cmd) and (
                    self.__note is None or self.__matches[cmd][OVERIDABLE]
                ):
                    self.__note = self.__matches[cmd]

        elif cmd == "required" and not self.__match(tags):
            self.__matches["required"] = (tags, arg, False)

    def __is_overidable(self, cmd: str) -> bool:
        return cmd not in self.__matches or self.__matches[cmd][OVERIDABLE]

    @classmethod
    def __is_all(cls, tag_list: List[str]) -> bool:
        return len(tag_list) == 1 and tag_list[0].lower() == "all"

    def __is_dead_cmd(self, cmd: str) -> bool:
        return (
            cmd == "dead"
            and "dead" in self.__matches
            and self.__matches["dead"][ARG] != "false"
        )

    def __match(self, tag_list: List[str]) -> bool:
        """Match tags against the system tags.

        True if all non-negated tags and none of the negated tags in the given
        list are present in system tags.
        """
        for tag in tag_list:
            if not tag.startswith("!"):
                # If tag is non-negated, it must be present in system tags
                if not (tag.lower() in self.system_tags):
                    return False
            else:
                # If tag is negated, it must be absent from system tags
                if tag[1:].lower() in self.system_tags:
                    return False
        return True

    def __parse_file(self, filename: Union[str, List[str]]) -> None:
        have_opt_data = False
        if isinstance(filename, list):
            for lineno, line in enumerate(filename, 1):
                self.__process_opt_line(line, lineno)
            have_opt_data = True
        elif os.path.isfile(filename):
            with open(filename, "r") as optfile:
                for lineno, line in enumerate(optfile, 1):
                    self.__process_opt_line(line, lineno)
            have_opt_data = True

        if have_opt_data:
            if "required" in self.__matches:
                self.__matches["dead"] = self.__matches["required"]
                self.is_dead = True
            elif self.__note is not None:
                self.is_dead = False
            elif self.__is_dead_cmd("dead"):
                self.is_dead = True
            else:
                self.is_dead = False

            if (
                self.__note is not None and self.__note[OVERIDABLE]
            ) or not self.__enable_note:
                self.__note = None

    def __str__(self) -> str:
        result = ""

        if self.is_dead:
            result += 'dead="{}"\n'.format(
                re.sub('"', '\\"', self.__matches["dead"][ARG])
            )
        else:
            for k in self.__matches:
                if k != "dead":
                    result += '{}="{}"\n'.format(
                        k, re.sub('"', '\\"', self.__matches[k][ARG])
                    )

            if self.__note is not None:
                result += 'activating_tag="{}"\n'.format(
                    re.sub('"', '\\"', ",".join(self.__note[TAGS]))
                )

        result = result.rstrip()
        return result


def eval_main(argv: Optional[List[str]] = None) -> None:
    parser = argparse.ArgumentParser(
        description="Evaluate test.opt against set of tags"
    )
    parser.add_argument(
        "opt_filename", help='The name of the "opt" file to read and evaluate'
    )
    parser.add_argument(
        "tags_list",
        metavar="TAG",
        nargs="*",
        help="A tag. For ease of use, this program also supports"
        " the use of a comma-separated list of tags.",
    )
    args = parser.parse_args(argv)

    # Build the system_tags list from the list of tags passed on
    # the command line. The tags which include a comma are considered
    # to be a comma-separated lists of tags, so split those into
    # sub-list of system tags.
    system_tags = []
    for tag in args.tags_list:
        system_tags.extend(tag.split(","))

    try:
        opt_result = OptFileParse(system_tags, args.opt_filename)
    except BadFormattingError as exc:
        print(str(exc))
        sys.exit(1)
    print(str(opt_result))


def check_syntax_main(argv: list[str] | None = None) -> None:
    parser = argparse.ArgumentParser(
        description="""
        Look for syntax errors in "opt" files.

        Print nothing and exit with status code 0 if no syntax error was found.
        Print error messages and exit with status code 1 otherwise.
        """
    )
    parser.add_argument(
        "filenames", nargs="+", help='The name of the "opt" file to parse.'
    )
    args = parser.parse_args(argv)

    has_errors = False
    for filename in args.filenames:
        try:
            OptFileParse([], filename)
        except BadFormattingError as exc:
            print(f"{filename}: {exc}")
            has_errors = True
    sys.exit(1 if has_errors else 0)


if __name__ == "__main__":  # no coverage
    eval_main()



from __future__ import annotations

from enum import Enum
import os.path
from typing import Any, Dict, List, NoReturn, Optional

from e3.testsuite import logger
from e3.testsuite.driver import TestDriver
from e3.testsuite.optfileparser import OptFileParse


class TestControlKind(Enum):
    """Control how to run (or not!) testcases."""

    NONE = 0
    """Run the test the regular way."""

    SKIP = 1
    """Do not run the testcase, setting it SKIP."""

    XFAIL = 2
    """
    Run the test the regular way. If its status is PASS, correct it to
    XPASS. If it succeeds, correct it to XFAIL. Leave its status unchanged in
    other cases.
    """


class TestControl:
    """Control the execution and analysis of a testcase."""

    def __init__(
        self,
        message: Optional[str] = None,
        skip: bool = False,
        xfail: bool = False,
    ) -> None:
        """Initialize a TestControl instance.

        :param message: Optional message to convey with the test status.
        :param skip: Whether to skip the test execution.
        :param xfail: Whether we expect the test to fail. If the test should be
            skipped, and xfailed, we consider it failed even though it did not
            run.
        """
        self.skip = skip
        self.xfail = xfail
        self.message = message

        # Not none iff created by AdaCoreLegacyTestControlCreator
        self.opt_results: Optional[Dict[str, Any]] = None


class TestControlCreator:
    """Abstract class to create test controls."""

    def create(self, driver: TestDriver) -> TestControl:
        """Create a TestControl instance for the given test driver.

        Raise a ValueError exception if the configuration is invalid.

        :param driver: Test driver for which we must parse the "control"
            configuration.
        """
        raise NotImplementedError


class YAMLTestControlCreator(TestControlCreator):
    """Create test controls from "test.yaml"'s "control" entries."""

    def __init__(self, condition_env: Optional[Dict[str, Any]] = None) -> None:
        """Initialize a YAMLTestControlCreator instance.

        :param condition_env: Environment to pass to condition evaluation in
            control entries. If None, use an empty dictionary.
        """
        self.condition_env = {} if condition_env is None else condition_env

    def create(self, driver: TestDriver) -> TestControl:
        # If there is a "test.opt" file while the YAML "control" entry
        # mechanism is in use, it probably means someone mistakenly wrote a
        # test.opt file that will not be interpreted: be helpful and warn about
        # it.
        if os.path.exists(driver.test_dir("test.opt")):
            logger.warning(
                '{}: "test.opt" file found whereas only "control" entries are'
                " considered".format(driver.test_env["test_name"])
            )

        # Read the configuration from the test environment's "control" key, if
        # present.
        default = TestControl()
        try:
            control = driver.test_env["control"]
        except KeyError:
            return default

        # Variables available to entry conditions
        condition_env = dict(self.condition_env)
        condition_env["env"] = driver.env

        # First validate the whole control structure, and only then interpret
        # it, for the same reason an language interpreter checks the syntax
        # before starting the interpretation.
        #
        # We expect control to be a list of lists of strings. The top-level
        # list is a collection of "entries": each entry conditionally selects a
        # test behavior. Each entry (list of strings) have one of the following
        # format:
        #
        #    [kind, condition]
        #    [kind, condition, message]
        #
        # "kind" is the name of any of the TestControlKind values. "condition"
        # is a Python expression that determines whether the entry applies, and
        # the optional "message" is a free form text to track which entry was
        # selected.
        entries = []

        if not isinstance(control, list):
            raise ValueError("list expected at the top level")

        for i, entry in enumerate(control, 1):

            def error(message: str, i: int = i) -> NoReturn:
                raise ValueError("entry #{}: {}".format(i, message))

            if (
                not isinstance(entry, list)
                or not len(entry) in (2, 3)
                or any(not isinstance(s, str) for s in entry)
            ):
                error("list of 2 or 3 strings expected")

            # Decode the test control kind
            try:
                kind = TestControlKind[entry[0]]
            except KeyError:
                error("invalid kind: {}".format(entry[0]))

            # Evaluate the condition
            try:
                cond = eval(entry[1], condition_env)
            except Exception as exc:
                error(
                    "invalid condition ({}): {}".format(
                        type(exc).__name__, exc
                    )
                )

            message = entry[2] if len(entry) > 2 else None

            entries.append((kind, cond, message))

        # Now, select the first entry whose condition is True. By default,
        # fallback to "default".
        for kind, cond, message in entries:
            if cond:
                skip, xfail = {
                    TestControlKind.NONE: (False, False),
                    TestControlKind.SKIP: (True, False),
                    TestControlKind.XFAIL: (False, True),
                }[kind]
                return TestControl(message, skip, xfail)
        return default


class AdaCoreLegacyTestControlCreator(TestControlCreator):
    """Create test controls for "test.opt"-based legacy AdaCore testsuites."""

    def default_script(self, driver: TestDriver) -> str:
        """Return the default test script filename.

        :param driver: Test driver for which we must parse the "control"
            configuration.
        """
        # Use "test.cmd" by default. If it does not exist while there is a
        # "test.py" or a "test.sh" file, use that instead.
        if not os.path.isfile(driver.test_dir("test.cmd")):
            if os.path.isfile(driver.test_dir("test.py")):
                return "test.py"
            elif os.path.isfile(driver.test_dir("test.sh")):
                return "test.sh"
        return "test.cmd"

    def default_opt_results(
        self, driver: TestDriver
    ) -> Dict[str, Optional[str]]:
        """Return the default options. test.opt files can override these.

        By default, a test is not DEAD, SKIP, nor XFAIL. Its execution timeout
        is 780 seconds. Test script is "test.cmd" and its output is compared
        against "test.out".

        :param driver: Test driver for which we must parse the "control"
            configuration.
        """
        return {
            "RLIMIT": "780",
            "DEAD": None,
            "XFAIL": None,
            "SKIP": None,
            "OUT": "test.out",
            "CMD": self.default_script(driver),
            "FILESIZE_LIMIT": None,
            "TIMING": None,
            "NOTE": None,
        }

    def __init__(
        self, system_tags: List[str], opt_filename: str = "test.opt"
    ) -> None:
        """Initialize a OptfileTestControlCreator instance.

        :param system_tags: Tags to forward to OptFileParse().
        :param opt_filename: Name of the file to parse, relative to the test
            directory.
        """
        self.opt_filename = opt_filename
        self.system_tags = system_tags

    def create(self, driver: TestDriver) -> TestControl:
        # If there is a "control" entry in the testcase's test.yaml file while
        # the optfile mechanism is in use, it probably means someone mistakenly
        # wrote a "control" entry that will not be interpreted: be helpful and
        # warn about it.
        if "control" in driver.test_env:
            logger.warning(
                '{}: "control" entry found in test.yaml whereas only test.opt'
                " files are considered".format(driver.test_env["test_name"])
            )

        # If it exists, parse the "test.opt" file in the test directory.
        # Create a dummy optfile otherwise.
        filename = driver.test_dir(self.opt_filename)
        optfile = (
            OptFileParse(self.system_tags, filename)
            if os.path.exists(filename)
            else OptFileParse(self.system_tags, [])
        )

        # Create a TestControl depending on the contents of the optfile
        message = None
        skip = False
        xfail = False
        opt_results = optfile.get_values(self.default_opt_results(driver))
        if opt_results["DEAD"] is not None:
            message = opt_results["DEAD"]
            skip = True
        elif opt_results["SKIP"] is not None:
            message = opt_results["SKIP"]
            skip = True
            xfail = True
        elif opt_results["XFAIL"] is not None:
            message = opt_results["XFAIL"]
            xfail = True
        result = TestControl(message, skip, xfail)

        # Store results in the TestControl instance, so that "driver" has
        # access to it.
        result.opt_results = opt_results

        return result



"""Tests for the "e3-testsuite-report" script."""

import os.path
import shutil
import xml.etree.ElementTree as ET

from e3.testsuite import Testsuite as Suite
from e3.testsuite.driver import BasicTestDriver as BasicDriver
from e3.testsuite.report.index import ReportIndex
from e3.testsuite.report.display import generate_report, main
from e3.testsuite.result import FailureReason, TestStatus as Status
from e3.testsuite.utils import ColorConfig

from .utils import create_report, create_result, run_testsuite


def run_status(results, argv, tmp_path, capsys):
    create_report(results, tmp_path)
    status = main(argv)
    captured = capsys.readouterr()
    assert not captured.err
    return status, captured.out


def run(results, argv, tmp_path, capsys, expect_failure=False):
    status, out = run_status(results, argv, tmp_path, capsys)
    if expect_failure:
        assert status != 0
    else:
        assert status == 0
    return out


basic_results = [
    create_result("append", Status.PASS, time=1.0),
    create_result(
        "sub", Status.XFAIL, msg="Unexpected result", log="3 - 2 = 0"
    ),
    create_result(
        "mult",
        Status.FAIL,
        msg="Unexpected result",
        log="3 * 2 = 5",
        time=123.0,
    ),
    create_result(
        "div",
        Status.FAIL,
        msg="Unexpected output",
        log="Running the div...",
        diff="diff\n-4\n+5",
    ),
]

sep_line = "-" * 79


class MyDriver(BasicDriver):
    return_status = Status.PASS

    def run(self, prev, slot):
        pass

    def analyze(self, prev, slot):
        self.result.set_status(Status.PASS)
        self.push_result()


class Mysuite(Suite):
    tests_subdir = "simple-tests"
    test_driver_map = {"default": MyDriver}

    @property
    def default_driver(self):
        return "default"


def test_basic(tmp_path, capsys):
    """Check that the script works fine in basic configurations."""
    assert run(basic_results, [str(tmp_path)], tmp_path, capsys) == (
        "Summary:\n"
        "\n"
        "  Out of 4 results\n"
        "  4 executed (not skipped)\n"
        "  PASS         1\n"
        "  FAIL         2\n"
        "  XFAIL        1\n"
        "\n"
        "  The following results may need further investigation:\n"
        "  2 new failure(s):\n"
        "    div: Unexpected output\n"
        "    mult: Unexpected result\n"
        "\n"
        "  1 expected failure(s):\n"
        "    sub: Unexpected result\n"
        "\n"
        "Result logs:\n"
        "\n"
        f"{sep_line}\n"
        "FAIL            div: Unexpected output\n"
        f"{sep_line}\n"
        "\n"
        "diff\n"
        "-4\n"
        "+5\n"
        "\n"
        f"{sep_line}\n"
        "FAIL            mult: Unexpected result\n"
        f"{sep_line}\n"
        "\n"
        "3 * 2 = 5\n"
        "\n"
    )


def test_empty(tmp_path, capsys):
    """Check the output for empty reports."""
    assert run([], [str(tmp_path)], tmp_path, capsys) == (
        "Summary:\n"
        "\n"
        "  Out of 0 results\n"
        "  0 executed (not skipped)\n"
        "  <no test result>\n"
        "\n"
        "Result logs:\n"
        "\n"
        "No relevant logs to display\n"
    )


def test_all(tmp_path, capsys):
    """Check that the --all option works as expected."""
    assert run(basic_results, [str(tmp_path), "--all"], tmp_path, capsys) == (
        "Summary:\n"
        "\n"
        "  Out of 4 results\n"
        "  4 executed (not skipped)\n"
        "  PASS         1\n"
        "  FAIL         2\n"
        "  XFAIL        1\n"
        "\n"
        "  The following results may need further investigation:\n"
        "  2 new failure(s):\n"
        "    div: Unexpected output\n"
        "    mult: Unexpected result\n"
        "\n"
        "  1 expected failure(s):\n"
        "    sub: Unexpected result\n"
        "\n"
        "Result logs:\n"
        "\n"
        f"{sep_line}\n"
        "FAIL            div: Unexpected output\n"
        f"{sep_line}\n"
        "\n"
        "diff\n"
        "-4\n"
        "+5\n"
        "\n"
        f"{sep_line}\n"
        "FAIL            mult: Unexpected result\n"
        f"{sep_line}\n"
        "\n"
        "3 * 2 = 5\n"
        "\n"
        f"{sep_line}\n"
        "XFAIL           sub: Unexpected result\n"
        f"{sep_line}\n"
        "\n"
        "3 - 2 = 0\n"
        "\n"
        f"{sep_line}\n"
        "PASS            append\n"
        f"{sep_line}\n"
        "\n"
        "<all logs are empty>\n"
        "\n"
    )


def test_time(tmp_path, capsys):
    """Check that the --show-time-info option works as expected."""
    assert run(
        basic_results, [str(tmp_path), "--show-time-info"], tmp_path, capsys
    ) == (
        "Summary:\n"
        "\n"
        "  Out of 4 results\n"
        "  4 executed (not skipped)\n"
        "  PASS         1\n"
        "  FAIL         2\n"
        "  XFAIL        1\n"
        "\n"
        "  The following results may need further investigation:\n"
        "  2 new failure(s):\n"
        "    div: Unexpected output\n"
        "    mult: Unexpected result\n"
        "\n"
        "  1 expected failure(s):\n"
        "    sub: Unexpected result\n"
        "\n"
        "Result logs:\n"
        "\n"
        f"{sep_line}\n"
        "FAIL            div: Unexpected output\n"
        f"{sep_line}\n"
        "\n"
        "diff\n"
        "-4\n"
        "+5\n"
        "\n"
        f"{sep_line}\n"
        "FAIL     02m03s mult: Unexpected result\n"
        f"{sep_line}\n"
        "\n"
        "3 * 2 = 5\n"
        "\n"
    )


def test_no_error_output(tmp_path, capsys):
    """Check that the --no-error-output option works as expected."""
    assert run(
        basic_results, [str(tmp_path), "--no-error-output"], tmp_path, capsys
    ) == (
        "Summary:\n"
        "\n"
        "  Out of 4 results\n"
        "  4 executed (not skipped)\n"
        "  PASS         1\n"
        "  FAIL         2\n"
        "  XFAIL        1\n"
        "\n"
        "  The following results may need further investigation:\n"
        "  2 new failure(s):\n"
        "    div: Unexpected output\n"
        "    mult: Unexpected result\n"
        "\n"
        "  1 expected failure(s):\n"
        "    sub: Unexpected result\n"
        "\n"
        "Result logs:\n"
        "\n"
        "FAIL            div: Unexpected output\n"
        "FAIL            mult: Unexpected result\n"
    )


def test_failure_reasons(tmp_path, capsys):
    """Check that we report stats about failure reasons."""
    results = [
        create_result(
            "crash", Status.FAIL, failure_reasons={FailureReason.CRASH}
        ),
        create_result(
            "diff", Status.FAIL, failure_reasons={FailureReason.DIFF}
        ),
        create_result(
            "diff-memcheck",
            Status.FAIL,
            failure_reasons={FailureReason.DIFF, FailureReason.MEMCHECK},
        ),
    ]

    assert run(
        results,
        [str(tmp_path), "--no-error-output"],
        str(tmp_path),
        capsys,
    ) == (
        "Summary:\n"
        "\n"
        "  Out of 3 results\n"
        "  3 executed (not skipped)\n"
        "  FAIL         3, including:\n"
        "    CRASH        1\n"
        "    MEMCHECK     1\n"
        "    DIFF         2\n"
        "\n"
        "  The following results may need further investigation:\n"
        "  3 new failure(s):\n"
        "    crash\n"
        "    diff\n"
        "    diff-memcheck\n"
        "\n"
        "Result logs:\n"
        "\n"
        "FAIL            crash\n"
        "FAIL            diff\n"
        "FAIL            diff-memcheck\n"
    )


def test_old_result(tmp_path, capsys):
    """Check that --old-result-dir works as expected."""
    old_results = [
        create_result("fail-to-pass", Status.FAIL),
        create_result("fail-to-xfail", Status.FAIL),
        create_result("fail-to-xpass", Status.FAIL),
        create_result("fail-to-fail", Status.FAIL),
        create_result("pass-to-fail", Status.PASS),
        create_result("pass-to-pass", Status.PASS),
        create_result("pass-to-xfail", Status.PASS),
        create_result("pass-to-xpass", Status.PASS),
        create_result("skip-to-skip", Status.SKIP),
        create_result("skip-to-pass", Status.SKIP),
        create_result("skip-to-fail", Status.SKIP),
        create_result("xfail-to-skip", Status.XFAIL),
        create_result("xfail-to-pass", Status.XFAIL),
        create_result("xfail-to-fail", Status.XFAIL),
        create_result("xfail-to-xfail", Status.XFAIL),
        create_result("removed", Status.PASS),
    ]
    new_results = [
        create_result("fail-to-pass", Status.PASS),
        create_result("fail-to-xfail", Status.XFAIL),
        create_result("fail-to-xpass", Status.XPASS),
        create_result("fail-to-fail", Status.FAIL),
        create_result("pass-to-fail", Status.FAIL),
        create_result("pass-to-pass", Status.PASS),
        create_result("pass-to-xfail", Status.XFAIL),
        create_result("pass-to-xpass", Status.XPASS),
        create_result("skip-to-skip", Status.SKIP),
        create_result("skip-to-pass", Status.PASS),
        create_result("skip-to-fail", Status.FAIL),
        create_result("xfail-to-skip", Status.SKIP),
        create_result("xfail-to-pass", Status.PASS),
        create_result("xfail-to-fail", Status.FAIL),
        create_result("xfail-to-xfail", Status.XFAIL),
        create_result("to-verify", Status.VERIFY),
        create_result("not-applicable-test", Status.NOT_APPLICABLE),
        create_result("error", Status.ERROR),
    ]

    # Create a report for old results
    tmp_path = str(tmp_path)
    old_result_dir = os.path.join(tmp_path, "old")
    os.mkdir(old_result_dir)
    create_report(old_results, old_result_dir)

    new_result_dir = os.path.join(tmp_path, "new")
    os.mkdir(new_result_dir)
    assert run(
        new_results,
        [
            new_result_dir,
            "--no-error-output",
            "--old-result-dir",
            old_result_dir,
        ],
        new_result_dir,
        capsys,
    ) == (
        "Summary:\n"
        "\n"
        "  Out of 18 results\n"
        "  16 executed (not skipped)\n"
        "  1 new skipped test(s)\n"
        "  1 removed test(s)\n"
        "\n"
        "  PASS         4\n"
        "  FAIL         4\n"
        "  XFAIL        3\n"
        "  XPASS        2\n"
        "  VERIFY       1\n"
        "  SKIP         2\n"
        "  NOT_APPLICABLE 1\n"
        "  ERROR        1\n"
        "\n"
        "  The following results may need further investigation:\n"
        "  3 new failure(s):\n"
        "    pass-to-fail\n"
        "    skip-to-fail\n"
        "    xfail-to-fail\n"
        "\n"
        "  1 already detected failure(s):\n"
        "    fail-to-fail\n"
        "\n"
        "  1 fixed failure(s):\n"
        "    fail-to-pass\n"
        "\n"
        "  3 expected failure(s):\n"
        "    fail-to-xfail\n"
        "    pass-to-xfail\n"
        "    xfail-to-xfail\n"
        "\n"
        "  2 unexpected passed test(s):\n"
        "    fail-to-xpass\n"
        "    pass-to-xpass\n"
        "\n"
        "  1 test(s) requiring additional verification:\n"
        "    to-verify\n"
        "\n"
        "  1 test(s) aborted due to unknown error:\n"
        "    error\n"
        "\n"
        "Result logs:\n"
        "\n"
        "FAIL            pass-to-fail\n"
        "FAIL            skip-to-fail\n"
        "FAIL            xfail-to-fail\n"
        "FAIL            fail-to-fail\n"
        "VERIFY          to-verify\n"
        "ERROR           error\n"
        "NOT_APPLICABLE        not-applicable-test\n"
    )


def test_generate_text_report(tmp_path):
    """Check TestsuiteCore's --generate-text-report option."""
    tmp_path = str(tmp_path)

    def check_report(expected):
        with open(os.path.join(tmp_path, "new", "report")) as f:
            assert f.read() == expected

    args = [
        "--generate-text-report",
        "--output-dir",
        tmp_path,
        "--rotate-output-dirs",
    ]

    # First run the testsuite with no previous results
    run_testsuite(Mysuite, args=args)
    check_report(
        "Summary:\n"
        "\n"
        "  Out of 2 results\n"
        "  2 executed (not skipped)\n"
        "  PASS         2\n"
        "\n"
        "Result logs:\n"
        "\n"
        "No relevant logs to display\n"
    )

    # Replace the "new" report with other data and then re-run the testsuite in
    # the same result dir, to exercize results comparison in the generated
    # report.
    new_result_dir = os.path.join(tmp_path, "new")
    shutil.rmtree(new_result_dir)
    os.mkdir(new_result_dir)
    create_report(
        [
            create_result("test1", Status.PASS),
            create_result("test2", Status.FAIL),
        ],
        new_result_dir,
    )

    run_testsuite(Mysuite, args=args)
    check_report(
        "Summary:\n"
        "\n"
        "  Out of 2 results\n"
        "  2 executed (not skipped)\n"
        "  0 new skipped test(s)\n"
        "  0 removed test(s)\n"
        "\n"
        "  PASS         2\n"
        "\n"
        "  The following results may need further investigation:\n"
        "  0 new failure(s):\n"
        "\n"
        "  1 fixed failure(s):\n"
        "    test2\n"
        "\n"
        "Result logs:\n"
        "\n"
        "No relevant logs to display\n"
    )


def test_auto_generate(tmp_path):
    """Check TestsuiteCore's auto_generate_text_report property."""

    def check_report(report_expected):
        assert (tmp_path / "new" / "report").exists() == report_expected

    class MyDerivedSuite(Mysuite):
        auto_generate_text_report = True

    args = [f"--output-dir={tmp_path}"]

    # First run the testsuite with no arg: we expect the text report to be
    # automatically generated.
    run_testsuite(MyDerivedSuite, args=args)
    check_report(True)
    shutil.rmtree(str(tmp_path / "new"))

    # Disable its generation on the command line and make sure it's not
    # generated
    run_testsuite(MyDerivedSuite, args=args + ["--no-generate-text-report"])
    check_report(False)


def test_failure_exit_code(tmp_path, capsys):
    """Check that --failure-exit-code works as expected."""

    def check(statuses, expected_exit_code):
        status, _ = run_status(
            [
                create_result(f"t{i}", status)
                for i, status in enumerate(statuses)
            ],
            ["--failure-exit-code=10", str(tmp_path)],
            tmp_path,
            capsys,
        )
        assert status == expected_exit_code

    check([Status.PASS, Status.PASS, Status.XFAIL, Status.SKIP], 0)
    check([Status.PASS, Status.FAIL], 10)
    check([Status.PASS, Status.ERROR], 10)
    check([Status.FAIL, Status.ERROR], 10)


def test_xunit_output(tmp_path, capsys):
    """Check that --xunit-output generates a report."""
    xunit_file = str(tmp_path / "xunit.xml")
    run(
        [create_result("foo", Status.PASS)],
        ["--xunit-output", xunit_file, str(tmp_path)],
        tmp_path,
        capsys,
    )

    # Just check that this produces a valid XML file
    ET.parse(xunit_file)


def test_xunit_name(tmp_path, capsys):
    """Check that --xunit-name has the desired effect."""
    xunit_file = str(tmp_path / "xunit.xml")
    testsuite_name = "mytestsuite"
    run(
        [create_result("foo", Status.PASS)],
        [
            "--xunit-output",
            xunit_file,
            "--xunit-name",
            testsuite_name,
            str(tmp_path),
        ],
        tmp_path,
        capsys,
    )

    xml = ET.parse(xunit_file)
    root = xml.getroot()

    # Check the root node
    assert root.tag == "testsuites"
    assert root.get("name") == testsuite_name

    # Check the invididual testsuites
    testsuites = xml.findall(".//testsuite")
    assert len(testsuites) != 0
    assert all(e.get("name") == testsuite_name for e in testsuites)

    # Check the invidiual tests
    tests = xml.findall(".//testcase")
    assert len(tests) != 0
    assert all(e.get("classname") == testsuite_name for e in tests)


def test_output_file(tmp_path, capsys):
    """Check that nothing is printed on stdout when output is a file."""
    create_report(
        [
            create_result(
                "foo", Status.FAIL, msg="Problem", log="This test failed"
            )
        ],
        tmp_path,
    )
    index = ReportIndex.read(str(tmp_path))
    filename = str(tmp_path / "report.txt")
    with open(filename, "w") as f:
        generate_report(
            output_file=f,
            new_index=index,
            old_index=None,
            colors=ColorConfig(colors_enabled=False),
            show_all_logs=True,
            show_xfail_logs=True,
            show_error_output=True,
            show_time_info=True,
        )

    captured = capsys.readouterr()
    assert captured.out == ""
    assert captured.err == ""

    with open(filename) as f:
        assert f.read() == (
            "Summary:\n"
            "\n"
            "  Out of 1 results\n"
            "  1 executed (not skipped)\n"
            "  FAIL         1\n"
            "\n"
            "  The following results may need further investigation:\n"
            "  1 new failure(s):\n"
            "    foo: Problem\n"
            "\n"
            "Result logs:\n"
            "\n"
            "-----------------------------------------------------------------"
            "--------------\n"
            "FAIL            foo: Problem\n"
            "-----------------------------------------------------------------"
            "--------------\n"
            "\n"
            "This test failed\n"
            "\n"
        )



"""Helpers to automatically rewrite test baseline."""

from __future__ import annotations

import abc
import dataclasses
import glob
import os.path
import sys

from e3.testsuite.report.index import ReportIndex
from e3.testsuite.result import FailureReason, TestStatus
from e3.testsuite.utils import ColorConfig


class RewritingError(Exception):
    """Raised by BaseBaselineRewriter.rewrite in case of fatal error."""

    pass


@dataclasses.dataclass
class RewritingSummary:
    """Summary of rewritten baselines."""

    errors: set[str]
    """
    Set of test names whose result was an error (baseline not updated).
    """

    updated_baselines: set[str]
    """Set of baselines that were updated.

    These are the baselines that changed, but that were not created nor
    deleted: they existed before the rewriting, they exist after and their
    contents are different.
    """

    new_baselines: set[str]
    """Set of test names whose baseline was created."""

    deleted_baselines: set[str]
    """Set of test names whose baseline file was removed."""


class BaseBaselineRewriter(abc.ABC):
    """Base class to rewrite test baselines from testsuite results."""

    def __init__(self, colors: ColorConfig, default_encoding: str = "utf-8"):
        """Baseline rewriter constructor.

        :param colors: Color configuration for messages printed to sys.stderr.
        :param default_encoding: Default encoding to use in order to write
            baseline files.
        """
        self.colors = colors
        self.default_encoding = default_encoding

    @abc.abstractmethod
    def baseline_filename(self, test_name: str) -> str:
        """
        Return the filename that contains the baseline for the given test.

        :param test_name: Name of the test for which we want the baseline
            filename.
        """
        ...

    def postprocess_baseline(self, baseline: bytes) -> bytes:
        """Refine a baseline to rewrite.

        By default, this returns the argument unchanged. Subclasses can
        override this if they need to refine baselines.
        """
        return baseline

    def rewrite(self, results_dir: str) -> RewritingSummary:
        """Rewrite baselines from a testsuite report.

        :param results_dir: Name of the directory in which to read the test
            results used to update baselines. That directory can contain either
            a native e3-testsuite report index, or a GAIA report.
        :return: A summary of tests that were processed.
        """
        summary = RewritingSummary(set(), set(), set(), set())

        # Try to detect the report format present in the results dir...

        # Try to read a "native" e3-testsuite report index
        if os.path.exists(
            os.path.join(results_dir, ReportIndex.INDEX_FILENAME)
        ):
            self.rewrite_from_index(summary, ReportIndex.read(results_dir))
            return summary

        # Try to read results from a GAIA report
        gaia_results = glob.glob(os.path.join(results_dir, "*.result"))
        if gaia_results:
            self.rewrite_from_gaia(summary, gaia_results)
            return summary

        raise RewritingError(
            f"could not find testsuite results in {results_dir}"
        )

    def rewrite_from_index(
        self, summary: RewritingSummary, index: ReportIndex
    ) -> None:
        """Rewrite baselines from a native e3-testsuite report index."""
        for test_name, entry in index.entries.items():
            # Only rewrite the output of (not expected) failed tests, and which
            # failed because of a DIFF
            if entry.status == TestStatus.FAIL:
                if entry.failure_reasons != {FailureReason.DIFF}:
                    self.handle_test_error(
                        summary,
                        test_name,
                        "test failed for another reason than a diff",
                    )
                    continue

                result = entry.load()

                # In case the output is a string, we need its encoding to write
                # it down to a file.
                assert result.env is not None
                encoding = result.env.get("encoding", self.default_encoding)

                if result.out is None:
                    self.handle_test_error(
                        summary,
                        test_name,
                        "no output associated to test result",
                    )
                    continue
                assert isinstance(result.out, (bytes, str))
                self.handle_test_diff(summary, test_name, result.out, encoding)

            elif entry.status == TestStatus.ERROR:
                self.handle_test_error(
                    summary, test_name, "test aborted because of an error"
                )

    def rewrite_from_gaia(
        self, summary: RewritingSummary, results: list[str]
    ) -> None:
        """Rewrite baselines from a GAIA report."""
        for filename in results:
            test_name = os.path.splitext(os.path.basename(filename))[0]

            with open(filename) as result_fp:
                # According to the GAIA format documentation, this file should
                # contain "TEST_STATUS:COMMENT", but in practice it contains
                # only a single letter for the status. Tolerate both.
                serialized_result = result_fp.read().strip()
                if len(serialized_result) == 1:
                    status = {
                        "O": "OK",
                        "I": "FAIL",
                        "X": "XFAIL",
                        "U": "UOK",
                        "T": "VERIFY",
                        "N": "DEAD",
                        "n": "NOT-APPLICABLE",
                        "C": "PROBLEM",  # Could also be CRASH/TIMEOUT
                        "D": "DIFF",
                    }[serialized_result]
                else:
                    status = serialized_result.split(":", 1)[0]

            # Consider that there is a diff if we get a DIFF ("D")
            if status == "DIFF":
                # Fetch the new baseline: look for an X.out file in the same
                # directory as "filename" (X.result). If not found, consider
                # that the new baseline is empty.
                new_baseline = b""
                out_file = os.path.join(
                    os.path.dirname(filename), f"{test_name}.out"
                )
                if os.path.exists(out_file):
                    with open(out_file, "rb") as baseline_fp:
                        new_baseline = baseline_fp.read()

                self.handle_test_diff(
                    summary, test_name, new_baseline, self.default_encoding
                )

            # Consider that the test aborted because of a fatal error if we get
            # a CRASH/PROBLEM/TIMEOUT ('C') or a simple FAIL.
            elif status == "PROBLEM":
                self.handle_test_error(
                    summary, test_name, "test status is CRASH/PROBLEM/TIMEOUT"
                )

    def handle_test_error(
        self, summary: RewritingSummary, test_name: str, reason: str
    ) -> None:
        """Notify users that a test result is an error."""
        summary.errors.add(test_name)
        self.print_warning(f"cannot update baseline for {test_name}: {reason}")

    def handle_test_diff(
        self,
        summary: RewritingSummary,
        test_name: str,
        new_baseline: bytes | str,
        encoding: str,
    ) -> None:
        """Rewrite the baseline of a single test."""
        filename = self.baseline_filename(test_name)
        baseline_exists = os.path.exists(filename)

        baseline_bytes = self.postprocess_baseline(
            new_baseline
            if isinstance(new_baseline, bytes)
            else new_baseline.encode(encoding)
        )
        if baseline_bytes:
            if baseline_exists:
                summary.updated_baselines.add(test_name)
            else:
                self.print_info(
                    f"no baseline file for {test_name}, creating it"
                )
                summary.new_baselines.add(test_name)
            with open(filename, "wb") as f:
                f.write(baseline_bytes)

        else:
            if baseline_exists:
                self.print_info(
                    f"baseline file found for {test_name}, deleting it"
                )
                summary.deleted_baselines.add(test_name)
                os.unlink(filename)

    def print_stderr(self, message: str, prefix: str, style: str = "") -> None:
        print(
            f"{style}{prefix}{self.colors.Style.RESET_ALL}: {message}",
            file=sys.stderr,
        )

    def print_info(self, message: str) -> None:
        self.print_stderr(message, "INFO")

    def print_warning(self, message: str) -> None:
        self.print_stderr(message, "WARNING", self.colors.Fore.YELLOW)

    def print_error(self, message: str) -> None:
        self.print_stderr(
            message,
            "ERROR",
            self.colors.Fore.RED + self.colors.Style.BRIGHT,
        )



"""Helpers for test drivers to spawn subprocesses.

Not only these lack documentation, these helpers as they are today are very
incomplete. For instance, it is not possible to disable relocated builds in
"gprbuild". So consider these as experimental and likely to break in the
future: use at your own risk.
"""

from __future__ import annotations

import os
from typing import Any, List, Optional

from e3.fs import mkdir
from e3.os.process import Run
from e3.testsuite import TestAbort
from e3.testsuite.driver import TestDriver
from e3.testsuite.result import Log, TestResult, TestStatus


def check_call(
    driver: TestDriver,
    cmd: List[str],
    test_name: Optional[str] = None,
    result: Optional[TestResult] = None,
    **kwargs: Any,
) -> Run:
    if "cwd" not in kwargs and "working_dir" in driver.test_env:
        kwargs["cwd"] = driver.test_env["working_dir"]
    if result is None:
        result = driver.result
    if test_name is None:
        test_name = driver.test_name
    process = Run(cmd, **kwargs)
    result.processes.append(
        {
            "output": Log(process.out),
            "status": process.status,
            "cmd": cmd,
            "run_args": kwargs,
        }
    )

    # Append the status code and process output to the log to ease post-mortem
    # investigation.
    result.log += "Status code: {}\n".format(process.status)
    result.log += "Output:\n"
    assert process.out is not None
    result.log += process.out

    if process.status != 0:
        result.set_status(TestStatus.FAIL, "command call fails")
        driver.push_result(result)
        raise TestAbort
    return process


def gprbuild(
    driver: TestDriver,
    project_file: Optional[str] = None,
    cwd: Optional[str] = None,
    gcov: bool = False,
) -> bool:
    if project_file is None:
        project_file = os.path.join(driver.test_env["test_dir"], "test.gpr")
    if cwd is None:
        cwd = driver.test_env["working_dir"]
    mkdir(cwd)
    gprbuild_cmd = [
        "gprbuild",
        "--relocate-build-tree",
        "-p",
        "-P",
        project_file,
    ]
    if gcov:
        gprbuild_cmd += ["-largs", "-lgcov"]
    check_call(driver, gprbuild_cmd, cwd=cwd)
    return True



"""Tests for e3.testsuite.optfileparser."""

import os.path

from e3.os.process import Run
from e3.testsuite.optfileparser import BadFormattingError, OptFileParse


optfiles_dir = os.path.join(os.path.dirname(__file__), "optfiles")


def parse_file(filename, tags=None):
    """Parse an optfile in the "optfiles" subdirectory."""
    tags = tags if tags is not None else []
    return OptFileParse(
        tags, os.path.join(os.path.dirname(__file__), "optfiles", filename)
    )


def parse(lines, tags=None):
    """Parse an optfile from its list of lines."""
    tags = tags if tags is not None else []
    return OptFileParse(tags, lines)


def test_in_memory():
    """Check the parsing of an optfile from a list of text lines."""
    of = parse_file("empty.opt")
    of = OptFileParse([], ["ALL"])
    assert of.get_values({}) == {}


def test_empty():
    """Test that parsing an empty file gives the expected values."""
    of = parse_file("empty.opt")
    assert of.get_value("XFAIL") is None
    assert of.get_values({}) == {}
    assert of.get_values({"XFAIL": "yes"}) == {"XFAIL": "yes"}


def test_tags():
    """Test the handling of system tags."""
    of = parse_file("tags.opt", [])
    assert of.get_value("CMD") == "default.cmd"
    assert of.get_value("XFAIL") == ""
    assert of.get_note() == ""
    assert of.get_note("") == []

    of = parse_file("tags.opt", ["linux"])
    assert of.get_value("CMD") == "linux.cmd"
    assert of.get_value("XFAIL") is None

    of = parse_file("tags.opt", ["powerpc"])
    assert of.get_value("CMD") == "default.cmd"
    assert of.get_value("XFAIL") == ""

    of = parse_file("tags.opt", ["linux", "powerpc"])
    assert of.get_value("CMD") == "linux.cmd"
    assert of.get_value("XFAIL") is None

    of = parse_file("tags.opt", ["aix", "powerpc"])
    assert of.get_value("CMD") == "aix.cmd"
    assert of.get_value("XFAIL") == ""

    # Tags can also be encoded as a comma-separated string
    of = parse_file("tags.opt", "vms,alpha")
    assert of.get_value("CMD") == "default.cmd"
    assert of.get_value("XFAIL") == ""


def test_dead():
    """Test the handling of DEAD entries."""
    of = parse_file("dead.opt")
    assert of.is_dead
    assert of.get_value("dead") == "true"
    assert of.get_value("xfail") is None
    assert of.get_note() == ""
    assert str(of) == 'dead="true"'

    of = parse_file("dead.opt", "linux")
    assert not of.is_dead
    assert of.get_value("dead") is None
    assert of.get_value("xfail") == ""
    assert of.get_value("cmd") is None
    assert of.get_note() == "linux"
    assert str(of) == 'xfail=""\nactivating_tag="linux"'

    of = parse_file("dead.opt", "darwin")
    assert not of.is_dead
    assert of.get_value("dead") is None
    assert of.get_value("xfail") == "Indeed!"
    assert of.get_value("cmd") is None
    assert of.get_note() == "darwin"

    of = parse_file("dead.opt", "windows")
    assert not of.is_dead
    assert of.get_value("dead") is None
    assert of.get_value("xfail") is None
    assert of.get_value("cmd") == "windows.cmd"
    assert of.get_note() == "windows"

    of = parse_file("dead.opt", "windows,powerpc")
    assert not of.is_dead
    assert of.get_value("dead") is None
    assert of.get_value("xfail") is None
    assert of.get_value("cmd") == "windows.cmd"
    assert of.get_note() == "windows"

    of = parse_file("dead.opt", "aix")
    assert of.is_dead
    assert of.get_value("dead") == "true"
    assert of.get_value("xfail") is None
    assert of.get_value("cmd") is None
    assert of.get_note() == ""

    of = parse_file("dead.opt", "powerpc,aix")
    assert not of.is_dead
    assert of.get_value("dead") is None
    assert of.get_value("xfail") is None
    assert of.get_value("cmd") == "aix-ppc.cmd"
    assert of.get_note() == "aix,powerpc"
    assert of.get_note("") == ["aix", "powerpc"]


def test_default_cmd():
    """Test the handling of default commands."""
    of = parse_file("default_cmd.opt", "linux")
    assert not of.is_dead

    of = parse_file("default_cmd.opt", "windows")
    assert of.is_dead


def test_syntax_error():
    """Test the handling of syntax errors."""
    try:
        parse_file("syntax_error.opt")
    except BadFormattingError as exc:
        assert str(exc) == "Can not parse line 2: ? ?"
    else:
        raise AssertionError()


def test_required():
    """Test the handling of REQUIRED commands."""
    of = parse_file("required.opt", "linux")
    assert of.is_dead

    of = parse_file("required.opt", "linux,ada")
    assert of.is_dead

    of = parse_file("required.opt", "linux,ada,c")
    assert not of.is_dead


def run_opt_parser_script(filename, tags=None):
    """Call e3-opt-parser and return the corresponding Run object."""
    parser_cmd = ["e3-opt-parser", os.path.join(optfiles_dir, filename)]
    if tags is not None:
        parser_cmd.extend(tags)
    return Run(parser_cmd)


def test_eval_main():
    """Test the function called by the command-line wrapper to OptFileParse."""
    p = run_opt_parser_script("tags.opt", None)
    assert p.status == 0
    assert (
        p.out
        == """\
cmd="default.cmd"
xfail=""
"""
    )

    p = run_opt_parser_script("tags.opt", ["linux"])
    assert p.status == 0
    assert (
        p.out
        == """\
cmd="linux.cmd"
"""
    )

    p = run_opt_parser_script("tags.opt", ["linux", "powerpc"])
    assert p.status == 0
    assert (
        p.out
        == """\
cmd="linux.cmd"
"""
    )


def test_eval_main_syntax_error():
    """Check that e3-opt-parser exits cleanly in case of parsing error."""
    p = run_opt_parser_script("syntax_error.opt", None)
    assert p.status == 1
    assert p.out == "Can not parse line 2: ? ?\n"


def test_check_syntax_main():
    """Check that e3-opt-check behaves as expected."""
    p = Run(
        [
            "e3-opt-check",
            "dead.opt",
            "syntax_error.opt",
            "tags.opt",
            "syntax_error_2.opt",
        ],
        cwd=optfiles_dir,
    )
    assert p.status == 1
    assert p.out == (
        "syntax_error.opt: Can not parse line 2: ? ?\n"
        "syntax_error_2.opt: Can not parse line 2:  ALL DEAD\n"
    )

    p = Run(["e3-opt-check", "dead.opt", "tags.opt"], cwd=optfiles_dir)
    assert p.status == 0
    assert p.out == ""


def test_sharp():
    """Check the handling for lines that contain '#'."""
    of = parse_file("sharp_in_comment.opt")
    assert of.get_value("dead") == "# This # can # contain # sharps"

    try:
        parse_file("syntax_error_3.opt")
    except BadFormattingError as exc:
        assert str(exc) == (
            "Can not parse line 1:"
            " # This is not a comment and should be rejected"
        )
    else:
        raise AssertionError()



from __future__ import annotations

import collections.abc
from dataclasses import dataclass
import os.path
import re
from typing import List, Optional, TYPE_CHECKING, Type, Union

from e3.env import Env
import e3.yaml

from e3.testsuite.driver import TestDriver

if TYPE_CHECKING:
    from e3.testsuite import TestsuiteCore


@dataclass
class ParsedTest:
    """Basic information to instantiate a test driver."""

    test_name: str
    """Name for this testcase."""

    driver_cls: Optional[Type[TestDriver]]
    """Test driver class to instantiate, None to use the default one."""

    test_env: dict
    """Base test environment.

    Driver instantiation will complete it with test directory, test name, etc.
    """

    test_dir: str
    """Directory that contains the testcase."""

    test_matcher: Optional[str] = None
    """Textual text matcher.

    If not None, string to match against the list of requested tests to run: in
    that case, the test is ignored if there is no match. This is needed to
    filter out tests in testsuites where tests don't necessarily have dedicated
    directories.
    """


TestFinderResult = Union[Optional[ParsedTest], List[ParsedTest]]


class ProbingError(Exception):
    """Exception raised in TestFinder.probe when a test is misformatted."""

    pass


class TestFinder:
    """Interface for objects that find testcases in the tests subdirectory."""

    @property
    def test_dedicated_directory(self) -> bool:
        """Return whether each test has a dedicated test directory.

        Even though e3-testsuite is primarily designed for this to be true,
        some testsuites actually host multiple tests in the same directory.
        When this is the case, we need to probe all directories and only then
        filter which test to run using ParsedTest.test_matcher.
        """
        return True

    def probe(
        self,
        testsuite: TestsuiteCore,
        dirpath: str,
        dirnames: List[str],
        filenames: List[str],
    ) -> TestFinderResult:
        """Return a test if the "dirpath" directory contains a testcase.

        Raise a ProbingError if anything is wrong.

        :param testsuite: Testsuite instance that is looking for testcases.
        :param dirpath: Directory to probe for a testcase.
        :param dirnames: List of directories that "dirpath" contains.
        :param filenames: List of files that "dirpath" contains.
        """
        raise NotImplementedError


class YAMLTestFinder(TestFinder):
    """
    Look for "test.yaml"-based tests.

    This considers that all directories that contain a "test.yaml" file are
    testcases. This file is parsed as YAML, the result is used as a test
    environment, and if it contains a "driver" key, it uses the testsuite
    driver whose name corresponds to the associated string value.
    """

    def probe(
        self,
        testsuite: TestsuiteCore,
        dirpath: str,
        dirnames: List[str],
        filenames: List[str],
    ) -> TestFinderResult:
        # There is a testcase iff there is a "test.yaml" file
        if "test.yaml" not in filenames:
            return None
        test_name = testsuite.test_name(dirpath)
        yaml_file = os.path.join(dirpath, "test.yaml")

        # Load the YAML file to build the test environment
        try:
            test_env = e3.yaml.load_with_config(yaml_file, Env().to_dict())
        except e3.yaml.YamlError as exc:
            raise ProbingError("invalid syntax for test.yaml") from exc

        # Ensure that the test_env act like a dictionary. We still accept None
        # as it's a shortcut for "just use default driver" configuration files.
        if test_env is None:
            test_env = {}
        elif not isinstance(test_env, collections.abc.Mapping):
            raise ProbingError("invalid format for test.yaml")

        driver_name = test_env.get("driver")
        if driver_name is None:
            driver_cls = None
        else:
            try:
                driver_cls = testsuite.test_driver_map[driver_name]
            except KeyError as exc:
                raise ProbingError("cannot find driver") from exc

        return ParsedTest(test_name, driver_cls, test_env, dirpath)


class AdaCoreLegacyTestFinder(TestFinder):
    """Look for testcases in directories whose name matches a Ticket Number."""

    TN_RE = re.compile("[0-9A-Z]{2}[0-9]{2}-[A-Z0-9]{3}")

    def __init__(self, driver_cls: Type[TestDriver]) -> None:
        """
        Initialize an AdaCoreLegacyTestFinder instance.

        :param driver_cls: TestDriver subclass to use for all tests that are
            found.
        """
        self.driver_cls = driver_cls

    def probe(
        self,
        testsuite: TestsuiteCore,
        dirpath: str,
        dirnames: List[str],
        filenames: List[str],
    ) -> TestFinderResult:
        # There is a testcase iff the test directory name is a valid TN
        dirname = os.path.basename(dirpath)
        if not self.TN_RE.match(dirname):
            return None

        return ParsedTest(dirname, self.driver_cls, {}, dirpath)



"""
Tests for the code e3.testsuite framework.

This checks the behavior of the Testsuite, TestDriver and BasicTestDriver
classes.
"""

import gc
import os
import re
import warnings
import weakref

from e3.fs import mkdir, mv
from e3.testsuite import TestAbort as E3TestAbort, Testsuite as Suite
from e3.testsuite.driver import (
    TestDriver as Driver,
    BasicTestDriver as BasicDriver,
)
from e3.testsuite.driver.diff import DiffTestDriver as DiffDriver
from e3.testsuite.report.index import ReportIndex
from e3.testsuite.result import (
    TestResult as Result,
    TestResultSummary as ResultSummary,
    TestStatus as Status,
)
from e3.testsuite.testcase_finder import (
    TestFinder as Finder,
    ParsedTest,
    YAMLTestFinder,
)

from .utils import (
    MultiSchedulingSuite,
    check_result,
    check_result_dirs,
    check_result_from_prefix,
    create_result,
    create_testsuite,
    extract_results,
    run_testsuite,
    run_testsuite_status,
    suite_logs,
)


class TestBasic:
    """Basic driver with all tests passing."""

    class MyDriver(BasicDriver):
        def run(self, prev, slot):
            pass

        def analyze(self, prev, slot):
            self.result.set_status(Status.PASS)
            self.push_result()

    class Mysuite(Suite):
        tests_subdir = "simple-tests"

        @property
        def test_driver_map(self):
            return {"default": TestBasic.MyDriver}

        @property
        def default_driver(self):
            return "default"

    def test(self):
        # Run the testsuite and check both the in-memory report and the on-disk
        # one.
        result = {"test1": Status.PASS, "test2": Status.PASS}
        suite = run_testsuite(self.Mysuite)
        assert extract_results(suite) == result
        check_result_dirs(new=result)


class TestOuterTestcase:
    """Check that we can run tests from another directory."""

    class MyDriver(BasicDriver):
        def run(self, prev, slot):
            pass

        def analyze(self, prev, slot):
            self.result.set_status(Status.PASS)
            self.push_result()

    class Mysuite(Suite):
        tests_subdir = "empty-tests"

        @property
        def test_driver_map(self):
            return {"default": TestOuterTestcase.MyDriver}

        default_driver = "default"

    def test(self):
        outer_test_dir = os.path.join(
            os.path.abspath(os.path.dirname(__file__)), "simple-tests"
        )
        suite = run_testsuite(self.Mysuite, args=[outer_test_dir])
        assert extract_results(suite) == {
            "simple-tests__test1": Status.PASS,
            "simple-tests__test2": Status.PASS,
        }


class TestInvalidFilterPattern:
    """Check the proper detection of invalid regexps on the command line."""

    class MyDriver(BasicDriver):
        def run(self, prev, slot):
            pass

        def analyze(self, prev, slot):
            self.result.set_status(Status.PASS)
            self.push_result()

    class Mysuite(Suite):
        tests_subdir = "simple-tests"

        @property
        def test_driver_map(self):
            return {"default": TestInvalidFilterPattern.MyDriver}

        default_driver = "default"

    def test(self, caplog):
        run_testsuite(self.Mysuite, args=["\\h"])
        assert any(
            message.startswith(
                "Test pattern is not a valid regexp, try to match it as-is: "
            )
            for message in suite_logs(caplog)
        )


class TestDumpEnviron:
    """Check that --dump-environ works (at least does not crash)."""

    class MyDriver(BasicDriver):
        return_status = Status.PASS

        def run(self, prev, slot):
            pass

        def analyze(self, prev, slot):
            self.result.set_status(self.return_status)
            self.push_result()

    class Mysuite(Suite):
        tests_subdir = "simple-tests"

        @property
        def test_driver_map(self):
            return {"default": TestDumpEnviron.MyDriver}

        default_driver = "default"

    def test(self):
        run_testsuite(self.Mysuite, args=["--dump-environ"])
        assert os.path.exists(os.path.join("out", "new", "environ.sh"))


class TestNoTestcase:
    """Testsuite run with no testcase."""

    class MyDriver(BasicDriver):
        def run(self, prev, slot):
            pass

        def analyze(self, prev, slot):
            self.result.set_status(Status.PASS, "ok!")
            self.push_result()

    class Mysuite(Suite):
        tests_subdir = "no-test"

        @property
        def test_driver_map(self):
            return {"default": TestNoTestcase.MyDriver}

        default_driver = "default"

    def test(self, caplog):
        suite = run_testsuite(self.Mysuite)
        logs = suite_logs(caplog)
        assert extract_results(suite) == {}
        assert any("<no test result>" in message for message in logs)


class TestAbort:
    """Check for if TestAbort work."""

    class MyDriver(BasicDriver):
        def run(self, prev, slot):
            raise E3TestAbort
            return "INVALID"

        def analyze(self, prev, slot):
            if prev.get("run") is None:
                self.result.set_status(Status.PASS, "ok!")
            else:
                self.result.set_status(Status.FAIL, "unexpected return value")

            self.push_result()

    class Mysuite(MultiSchedulingSuite):
        tests_subdir = "simple-tests"

        @property
        def test_driver_map(self):
            return {"default": TestAbort.MyDriver}

        @property
        def default_driver(self):
            return "default"

    def run_check(self, multiprocessing):
        suite = run_testsuite(self.Mysuite, multiprocessing=multiprocessing)
        assert extract_results(suite) == {
            "test1": Status.PASS,
            "test2": Status.PASS,
        }

    def test_multithread(self):
        self.run_check(multiprocessing=False)

    def test_multiprocess(self):
        self.run_check(multiprocessing=True)


class TestExceptionInDriver:
    """Check handling of exception in test driver."""

    class MyDriver(BasicDriver):
        def run(self, prev, slot):
            raise AttributeError("expected exception")

        def analyze(self, prev, slot):
            # In multiprocessing mode, there is no return value propagation, so
            # there is nothing to check here.
            if self.env.use_multiprocessing:
                self.result.set_status(Status.PASS, "ok!")
            else:
                self.result.log += f"Previous values: {prev}\n"
                prev_value = prev["run"]
                if isinstance(prev_value, Exception):
                    self.result.set_status(Status.PASS, "ok!")
                else:
                    self.result.set_status(
                        Status.FAIL, "unexpected return value"
                    )
            self.push_result()

    class Mysuite(MultiSchedulingSuite):
        tests_subdir = "simple-tests"

        @property
        def test_driver_map(self):
            return {"default": TestExceptionInDriver.MyDriver}

        @property
        def default_driver(self):
            return "default"

    def run_check(self, multiprocessing):
        suite = run_testsuite(
            self.Mysuite, multiprocessing=multiprocessing, expect_failure=True
        )

        results = extract_results(suite)

        # Expect PASS for both tests
        assert results.pop("test1") == Status.PASS
        assert results.pop("test2") == Status.PASS

        # Expect two extra ERROR results for errors in MyDriver.run. Their
        # names depend on a counter we don't control, hence the involved
        # checking code.
        assert len(results) == 2

        keys = sorted(results)
        assert keys[0] == "test1.run"
        assert keys[1] == "test2.run"
        assert set(results.values()) == {Status.ERROR}

    def test_multithread(self):
        self.run_check(multiprocessing=False)

    def test_multiprocess(self):
        self.run_check(multiprocessing=True)


class TestNotExistingTempDir:
    """Check the detection of missing requested temporary directory."""

    class MyDriver(BasicDriver):
        def run(self, prev, slot):
            return True

        def analyze(self, prev, slot):
            self.result.set_status(Status.PASS)
            self.push_result()

    class Mysuite(Suite):
        tests_subdir = "simple-tests"

        @property
        def test_driver_map(self):
            return {"default": TestNotExistingTempDir.MyDriver}

        default_driver = "default"

    def test(self, caplog):
        # Check that the testsuite aborted and that we have the expected error
        # message.
        run_testsuite(self.Mysuite, ["--temp-dir=tmp"], expect_failure=True)
        logs = suite_logs(caplog)
        assert any(
            re.search("temp dir '.*tmp' does not exist", entry)
            for entry in logs
        )


class TestNoRandomTestDir:
    """Check support for --no-random-temp-subdir."""

    class MyDriver(BasicDriver):
        def run(self, prev, slot):
            mkdir(self.working_dir())
            return True

        def analyze(self, prev, slot):
            self.result.set_status(Status.PASS)
            self.push_result()

    class Mysuite(Suite):
        tests_subdir = "simple-tests"

        @property
        def test_driver_map(self):
            return {"default": TestNoRandomTestDir.MyDriver}

        default_driver = "default"

    def test(self, tmp_path):
        run_testsuite(
            self.Mysuite,
            [
                f"--temp-dir={tmp_path}",
                "--no-random-temp-subdir",
                "--cleanup-mode=none",
            ],
        )
        assert os.path.isdir(str(tmp_path / "test1"))
        assert os.path.isdir(str(tmp_path / "test2"))


class TestDevMode:
    """Check the dev mode (--dev-temp) works as expected."""

    class MyDriver(BasicDriver):
        def run(self, prev, slot):
            os.mkdir(self.test_env["working_dir"])
            path = os.path.join(self.test_env["working_dir"], "foo.txt")
            with open(path, "w") as f:
                f.write(self.test_env["test_name"])
            return True

        def analyze(self, prev, slot):
            self.result.set_status(Status.PASS)
            self.push_result()

    class Mysuite(Suite):
        tests_subdir = "simple-tests"

        @property
        def test_driver_map(self):
            return {"default": TestDevMode.MyDriver}

        default_driver = "default"

    def test(self):
        suite = run_testsuite(self.Mysuite, ["--dev-temp=tmp"])

        # Check testsuite report
        assert extract_results(suite) == {
            "test1": Status.PASS,
            "test2": Status.PASS,
        }

        # Check the presence and content of working directories
        for test in ["test1", "test2"]:
            path = os.path.join("tmp", test, "foo.txt")
            with open(path, "r") as f:
                content = f.read()
            assert content == test


class TestInvalidYAML:
    """Check that invalid test.yaml files are properly reported."""

    class MyDriver(BasicDriver):
        def run(self, prev, slot):
            pass

        def analyze(self, prev, slot):
            self.result.set_status(Status.PASS, "all good")
            self.push_result()

    class Mysuite(Suite):
        tests_subdir = "invalid-yaml-tests"

        @property
        def test_driver_map(self):
            return {"default": TestInvalidYAML.MyDriver}

        default_driver = "default"

    def test(self):
        # The testsuite is supposed to run to completion (valid tests have
        # run), but it ends with an error status code.
        suite = run_testsuite(self.Mysuite, expect_failure=True)
        results = suite.report_index.entries

        assert len(results) == 4
        assert results["valid"].status == Status.PASS

        check_result(
            suite,
            "invalid_syntax",
            Status.ERROR,
            "invalid syntax for test.yaml",
        )
        check_result(
            suite,
            "invalid_structure",
            Status.ERROR,
            "invalid format for test.yaml",
        )
        check_result(
            suite, "invalid_driver", Status.ERROR, "cannot find driver"
        )


class TestMissingDriver:
    """Check that missing drivers in test.yaml files are properly reported."""

    class MyDriver(BasicDriver):
        def run(self, prev, slot):
            pass

        def analyze(self, prev, slot):
            self.result.set_status(Status.PASS, "all good")
            self.push_result()

    class Mysuite(Suite):
        tests_subdir = "invalid-yaml-tests"

        @property
        def test_driver_map(self):
            return {"default": TestMissingDriver.MyDriver}

    def test(self):
        suite = run_testsuite(self.Mysuite, expect_failure=True)
        check_result(suite, "valid", Status.ERROR, "missing test driver")


class TestInvalidDriver:
    """Check that faulty driver classes are properly reported."""

    class MyDriver(BasicDriver):
        def __init__(self, *args, **kwargs):
            raise NotImplementedError("__init__ not implemented")

        def analyze(self):
            raise NotImplementedError

        def run(self):
            raise NotImplementedError

    class Mysuite(Suite):
        tests_subdir = "simple-tests"
        default_driver = "default"

        @property
        def test_driver_map(self):
            return {"default": TestInvalidDriver.MyDriver}

    def test(self):
        suite = run_testsuite(self.Mysuite, expect_failure=True)
        check_result(suite, "test1", Status.ERROR, "__init__ not implemented")


class TestDuplicateName:
    """Check duplicate test names are properly reported."""

    class MyDriver(BasicDriver):
        def run(self, prev, slot):
            self.result.log += "Work is being done..."

        def analyze(self, prev, slot):
            self.result.set_status(Status.PASS, "all good")
            self.push_result()

    class Mysuite(Suite):
        tests_subdir = "simple-tests"
        default_driver = "default"

        @property
        def test_driver_map(self):
            return {"default": TestDuplicateName.MyDriver}

        def test_name(self, test_dir):
            return "foo"

    def test(self):
        suite = run_testsuite(self.Mysuite, expect_failure=True)
        assert len(suite.report_index.entries) == 2
        check_result(suite, "foo", Status.ERROR, "duplicate test name: foo")
        check_result_from_prefix(
            suite,
            "foo__except",
            Status.ERROR,
            "cannot push twice results for foo",
        )


class TestShowErrorOutput:
    """Check that --show-error-output works as expected."""

    class MyDriver(Driver):
        def add_test(self, dag):
            self.add_fragment(dag, "run")

        def run(self, prev, slot):
            self.result.log += "Work is being done..."

            if self.test_env["test_name"] == "test1":
                self.result.set_status(Status.PASS, "all good")
            else:
                self.result.set_status(Status.FAIL, "test always fail!")

            self.push_result()

    class Mysuite(Suite):
        tests_subdir = "simple-tests"
        default_driver = "default"

        @property
        def test_driver_map(self):
            return {"default": TestShowErrorOutput.MyDriver}

    def test(self, caplog):
        run_testsuite(
            self.Mysuite, ["--show-error-output"], expect_failure=True
        )
        logs = suite_logs(caplog)
        assert any("Work is being done" in message for message in logs)


class TestPushTwice:
    """Test error detection when pushing results twice in a driver."""

    class MyDriver(BasicDriver):
        def run(self, prev, slot):
            pass

        def analyze(self, prev, slot):
            self.result.set_status(Status.PASS, "ok!")
            self.push_result()
            self.push_result()

    class Mysuite(Suite):
        tests_subdir = "simple-tests"

        @property
        def test_driver_map(self):
            return {"default": TestPushTwice.MyDriver}

        @property
        def default_driver(self):
            return "default"

    def test(self):
        suite = run_testsuite(
            self.Mysuite, args=["simple-tests/test1"], expect_failure=True
        )
        assert len(suite.report_index.entries) == 2
        check_result(
            suite,
            "test1__except1",
            Status.ERROR,
            "cannot push twice results for test1",
        )


def test_result_set_status_twice(caplog):
    """Test that calling TestResult.set_status twice is rejected."""
    r = Result("foobar")
    r.set_status(Status.PASS)
    r.set_status(Status.PASS)

    logs = {r.getMessage() for r in caplog.records}
    assert "cannot set test foobar status twice" in logs


def test_result_str(caplog):
    """Test that calling TestResult.set_status twice is rejected."""
    r = Result("foobar", msg="<message>")
    assert str(r) == "foobar                   TestStatus.ERROR <message>"


class TestDefaultCommentFile:
    """Test that the comment file is written as expected by default."""

    class MyDriver(BasicDriver):
        def run(self, prev, slot):
            pass

        def analyze(self, prev, slot):
            self.result.set_status(Status.PASS)
            self.push_result()

    class Mysuite(Suite):
        tests_subdir = "simple-tests"
        default_driver = "default"

        @property
        def test_driver_map(self):
            return {"default": TestDefaultCommentFile.MyDriver}

    def test(self):
        run_testsuite(self.Mysuite)
        with open(os.path.join("out", "new", "comment")) as f:
            content = f.readlines()
        assert len(content) == 2
        assert content[0] == "Testsuite options:\n"


class TestCommentFile:
    """Test that the comment file is written as expected."""

    class MyDriver(BasicDriver):
        def run(self, prev, slot):
            pass

        def analyze(self, prev, slot):
            self.result.set_status(Status.PASS)
            self.push_result()

    class Mysuite(Suite):
        tests_subdir = "simple-tests"
        default_driver = "default"

        @property
        def test_driver_map(self):
            return {"default": TestCommentFile.MyDriver}

        def write_comment_file(self, f):
            counters = self.report_index.status_counters
            lines = sorted(
                "{}:{}".format(status.name, counter)
                for status, counter in counters.items()
                if counter
            )
            f.write(" ".join(lines))

    def test(self):
        run_testsuite(self.Mysuite)
        with open(os.path.join("out", "new", "comment")) as f:
            content = f.read()
        assert content == "PASS:2"


class TestPathBuilders:
    """Check that path building methods in TestDriver work as expected."""

    class MyDriver(BasicDriver):
        def run(self, prev, slot):
            assert self.working_dir("foo.txt") == os.path.join(
                self.test_env["working_dir"], "foo.txt"
            )
            assert self.test_dir("foo.txt") == os.path.join(
                self.test_env["test_dir"], "foo.txt"
            )

        def analyze(self, prev, slot):
            self.result.set_status(Status.PASS)
            self.push_result()

    class Mysuite(Suite):
        tests_subdir = "simple-tests"
        default_driver = "default"

        @property
        def test_driver_map(self):
            return {"default": TestPathBuilders.MyDriver}

    def test(self):
        run_testsuite(self.Mysuite)


class TestMultilineMessage:
    """Check that multiline messages are adjusted in test results."""

    class MyDriver(BasicDriver):
        def run(self, prev, slot):
            pass

        def analyze(self, prev, slot):
            self.result.set_status(
                Status.PASS, "   Ugly  \nmultiline\r\n   \tstring  "
            )
            self.push_result()

    class Mysuite(Suite):
        tests_subdir = "simple-tests"

        @property
        def test_driver_map(self):
            return {"default": TestMultilineMessage.MyDriver}

        @property
        def default_driver(self):
            return "default"

    def tests(self):
        suite = run_testsuite(self.Mysuite, args=["test1"])
        result = suite.report_index.entries["test1"].load()
        assert result.status == Status.PASS
        assert result.msg == "Ugly multiline string"


class TestFailureExitCode:
    """Check that --failure-exit-code works as expected."""

    class MyDriver(BasicDriver):
        def run(self, prev, slot):
            pass

        def analyze(self, prev, slot):
            self.result.set_status(
                Status.PASS if self.test_name == "test1" else Status.FAIL
            )
            self.push_result()

    class Mysuite(Suite):
        tests_subdir = "simple-tests"
        default_driver = "default"

        @property
        def test_driver_map(self):
            return {"default": TestFailureExitCode.MyDriver}

    class Mysuite2(Mysuite):
        default_failure_exit_code = 2

    def check(self, cls, args, expected_status):
        suite, status = run_testsuite_status(cls, args)
        assert extract_results(suite) == {
            "test1": Status.PASS,
            "test2": Status.FAIL,
        }
        assert status == expected_status

    def test(self):
        self.check(self.Mysuite, ["--failure-exit-code=0"], 0)
        self.check(self.Mysuite, [], 1)
        self.check(self.Mysuite, ["--failure-exit-code=1"], 1)
        self.check(self.Mysuite2, [], 2)


class TestMaxConsecutiveFailures:
    """Check that --max-consecutive-failures works as expected."""

    class MyDriver(Driver):
        def add_test(self, dag):
            self.add_fragment(dag, "run")

        def run(self, prev, slot):
            self.result.set_status(Status.FAIL)
            self.push_result()

    def run_check(self, caplog, n_failures, multiprocessing, jobs=1):

        def adjust_dag_deps(suite, dag):
            # Make sure that n_failure tests run first, and that the other
            # tests start only after the n_failure tests completed.
            fragments = list(dag.vertex_data.values())
            first_batch = [f.uid for f in fragments[:n_failures]]
            second_batch = fragments[n_failures:]
            for f in second_batch:
                dag.update_vertex(vertex_id=f.uid, predecessors=first_batch)

        suite = run_testsuite(
            create_testsuite(
                ["a", "b", "c", "d"],
                self.MyDriver,
                MultiSchedulingSuite,
                adjust_dag_deps,
            ),
            args=[
                f"--max-consecutive-failures={n_failures}",
                f"-j{jobs}",
                "-E",
            ],
            multiprocessing=multiprocessing,
            expect_failure=True,
        )
        logs = {r.getMessage() for r in caplog.records}

        results = len(suite.report_index.entries)
        skip = 0
        failures = 0
        for e in suite.report_index.entries.values():
            if e.status == Status.FAIL:
                failures += 1
            else:
                assert e.status == Status.SKIP
                skip += 1
        assert n_failures == failures
        assert results - failures == skip
        assert "Too many consecutive failures, aborting the testsuite" in logs

    def test_multithread(self, caplog):
        self.run_check(caplog, n_failures=1, multiprocessing=False)
        self.run_check(caplog, n_failures=2, multiprocessing=False)

    def test_multiprocess(self, caplog):
        self.run_check(caplog, n_failures=2, multiprocessing=True)
        self.run_check(caplog, n_failures=2, multiprocessing=True, jobs=4)


class TestMaxConsecutiveFailuresBatch:
    """Check that --max-consecutive-failures works as expected for batches.

    Make a single test fragment push a batch of failures that triggers
    testsuite abortion.
    """

    class MyDriver(Driver):
        def add_test(self, dag):
            self.add_fragment(dag, "run")

        def run(self, prev, slot):
            for i in range(4):
                self.push_result(create_result(f"t{i}", Status.FAIL))

    def run_check(self, caplog, multiprocessing):
        suite = run_testsuite(
            create_testsuite(["t"], self.MyDriver, MultiSchedulingSuite),
            args=["--max-consecutive-failures=2"],
            multiprocessing=multiprocessing,
            expect_failure=True,
        )
        assert extract_results(suite) == {
            "t0": Status.FAIL,
            "t1": Status.FAIL,
            "t2": Status.FAIL,
            "t3": Status.FAIL,
        }
        logs = {r.getMessage() for r in caplog.records}
        assert "Too many consecutive failures, aborting the testsuite" in logs

    def test_multithread(self, caplog):
        self.run_check(caplog, multiprocessing=False)

    def test_multiprocess(self, caplog):
        self.run_check(caplog, multiprocessing=True)


class TestShowTimeInfo:
    """Check that --show-time-info works as expected."""

    class MyDriver(BasicDriver):
        def run(self, prev, slot):
            pass

        def analyze(self, prev, slot):
            self.result.time = 1.0
            self.result.set_status(Status.PASS)
            self.push_result()

    class Mysuite(Suite):
        tests_subdir = "simple-tests"
        default_driver = "default"

        @property
        def test_driver_map(self):
            return {"default": TestShowTimeInfo.MyDriver}

    def test(self, caplog):
        suite = run_testsuite(self.Mysuite, args=["--show-time-info", "test1"])
        logs = {r.getMessage() for r in caplog.records}
        test_summaries = [line for line in logs if line.startswith("PASS")]
        assert len(suite.report_index.entries) == 1
        assert test_summaries == ["PASS     00m01s test1"]


class TestDeprecated:
    """Test deprecated methods."""

    class MyDriver(BasicDriver):
        def run(self, prev, slot):
            pass

        def analyze(self, prev, slot):
            self.result.set_status(Status.PASS)
            self.push_result()

    class Mysuite(Suite):
        tests_subdir = "simple-tests"

        @property
        def test_driver_map(self):
            return {"default": TestDeprecated.MyDriver}

        @property
        def default_driver(self):
            return "default"

    def check_warning(self, warn_list):
        assert len(warn_list) == 1
        w = warn_list[0]
        assert issubclass(w.category, DeprecationWarning)
        assert "obsolete" in str(w.message)

    def test(self):
        suite = run_testsuite(self.Mysuite)

        with warnings.catch_warnings(record=True) as w:
            assert suite.test_counter == 2
            self.check_warning(w)

        with warnings.catch_warnings(record=True) as w:
            expected = {s: 0 for s in Status}
            expected[Status.PASS] = 2
            assert suite.test_status_counters == expected
            self.check_warning(w)

        with warnings.catch_warnings(record=True) as w:
            assert suite.results == {
                "test1": Status.PASS,
                "test2": Status.PASS,
            }
            self.check_warning(w)


class TestReadReportIndex:
    """Check that reading a report index works as expected."""

    class MyDriver(BasicDriver):
        def run(self, prev, slot):
            pass

        def analyze(self, prev, slot):
            self.result.set_status(Status.PASS)
            self.push_result()

    class Mysuite(Suite):
        tests_subdir = "simple-tests"

        @property
        def test_driver_map(self):
            return {"default": TestReadReportIndex.MyDriver}

        @property
        def default_driver(self):
            return "default"

    def test(self):
        run_testsuite(self.Mysuite)
        index = ReportIndex.read(os.path.join("out", "new"))
        summaries = {
            key: entry.summary for key, entry in index.entries.items()
        }
        assert summaries == {
            "test1": ResultSummary(
                "test1", Status.PASS, None, set(), None, {}
            ),
            "test2": ResultSummary(
                "test2", Status.PASS, None, set(), None, {}
            ),
        }


class TestMultipleTestsPerDir:
    """Test a test finder that returns multiple tests per directory."""

    class CustomTestFinder(Finder):
        test_dedicated_directory = False

        def probe(self, testsuite, dirpath, dirnames, filenames):
            result = []
            for f in filenames:
                if not f.endswith(".txt"):
                    continue

                # Strip the "*.txt" extension for the test name, but preserve
                # it for the matcher.
                test_name = testsuite.test_name(os.path.join(dirpath, f[:-4]))
                test_matcher = os.path.join(dirpath, f)

                result.append(
                    ParsedTest(
                        test_name=test_name,
                        driver_cls=TestMultipleTestsPerDir.MyDriver,
                        test_env={},
                        test_dir=dirpath,
                        test_matcher=test_matcher,
                    )
                )
            return result

    class MyDriver(BasicDriver):
        def run(self, prev, slot):
            pass

        def analyze(self, prev, slot):
            self.result.set_status(Status.PASS)
            self.push_result()

    class Mysuite(Suite):
        tests_subdir = "txt-tests"

        @property
        def test_driver_map(self):
            return {"default": TestMultipleTestsPerDir.MyDriver}

        @property
        def test_finders(self):
            return [TestMultipleTestsPerDir.CustomTestFinder()]

        @property
        def default_driver(self):
            return "default"

    def test(self):
        # Check a full testsuite run
        suite = run_testsuite(self.Mysuite)
        assert extract_results(suite) == {
            "bar__x": Status.PASS,
            "bar__y": Status.PASS,
            "foo__a": Status.PASS,
            "foo__b": Status.PASS,
            "foo__c": Status.PASS,
        }

        # Check filtering
        suite = run_testsuite(self.Mysuite, args=["a.txt"])
        assert extract_results(suite) == {"foo__a": Status.PASS}


class TestInterTestDeps:
    """Check we can run a testsuite with inter-tests dependencies."""

    # Run a testsuite with two kind of drivers: UnitDriver ones, which just
    # "compute a number" and SumDriver ones, which compute the sum of all
    # numbers from UnitDriver tests. SumDriver tests depend on the UnitDriver:
    # they must run after them and access their data.

    @staticmethod
    def result_filename(driver, unit_test_name):
        return os.path.join(
            driver.env.working_dir,
            f"result-{unit_test_name}.txt",
        )

    class UnitDriver(Driver):
        def add_test(self, dag):
            self.add_fragment(dag, "run")

        def run(self, prev, slot):
            result = int(self.test_name.split("_")[-1])
            with open(
                TestInterTestDeps.result_filename(self, self.test_name), "w"
            ) as f:
                f.write(str(result))

            self.result.set_status(Status.PASS)
            self.push_result()

    class SumDriver(Driver):
        def add_test(self, dag):
            self.add_fragment(dag, "run")

        def run(self, prev, slot):
            sum_result = 0

            for unit_test_name in self.test_env["unit_names"]:
                with open(
                    TestInterTestDeps.result_filename(self, unit_test_name)
                ) as f:
                    sum_result += int(f.read())

            self.result.set_status(
                Status.PASS if sum_result == 10 else Status.FAIL
            )
            self.push_result()

    class Mysuite(MultiSchedulingSuite):
        tests_subdir = "."

        @staticmethod
        def parsed_test(driver_cls, test_name):
            return ParsedTest(test_name, driver_cls, {}, ".")

        def get_test_list(self, sublist):
            return [
                self.parsed_test(TestInterTestDeps.SumDriver, "sum"),
                self.parsed_test(TestInterTestDeps.UnitDriver, "unit_0"),
                self.parsed_test(TestInterTestDeps.UnitDriver, "unit_1"),
                self.parsed_test(TestInterTestDeps.UnitDriver, "unit_2"),
                self.parsed_test(TestInterTestDeps.UnitDriver, "unit_3"),
                self.parsed_test(TestInterTestDeps.UnitDriver, "unit_4"),
            ]

        def adjust_dag_dependencies(self, dag):
            # Get the list of all fragments for...

            # ... UnitDriver.run
            unit_fragments = []

            # ... SumDriver.run
            sum_fragments = []

            for fg in dag.vertex_data.values():
                if fg.matches(TestInterTestDeps.UnitDriver, "run"):
                    unit_fragments.append(fg)
                elif fg.matches(TestInterTestDeps.SumDriver, "run"):
                    sum_fragments.append(fg)

            # Pass the list of UnitDriver.run fragments to all SumDriver
            # instances and make sure SumDriver fragments run after all
            # UnitDriver.run ones.
            unit_uids = [fg.uid for fg in unit_fragments]
            unit_names = [fg.driver.test_name for fg in unit_fragments]
            for fg in sum_fragments:
                fg.driver.test_env["unit_names"] = unit_names
                dag.update_vertex(vertex_id=fg.uid, predecessors=unit_uids)

    def run_check(self, multiprocessing):
        suite = run_testsuite(self.Mysuite, multiprocessing=multiprocessing)
        assert extract_results(suite) == {
            "unit_0": Status.PASS,
            "unit_1": Status.PASS,
            "unit_2": Status.PASS,
            "unit_3": Status.PASS,
            "unit_4": Status.PASS,
            "sum": Status.PASS,
        }

    def test_multithread(self):
        self.run_check(multiprocessing=False)

    def test_multiprocess(self):
        self.run_check(multiprocessing=True)


def test_results_relocation(tmp_path):
    """Check that results stored on the filesystem are relocatable."""
    old_dir = str(tmp_path / "old")
    new_dir = str(tmp_path / "new")

    # Create a report in "old_dir"
    mkdir(old_dir)
    old_index = ReportIndex(old_dir)
    for r in [
        Result("foo", status=Status.PASS),
        Result("bar", status=Status.FAIL),
    ]:
        old_index.add_result(r.summary, r.save(old_dir))
    old_index.write()

    # Move it to "new_dir", and check that we can still load its entries
    mv(old_dir, new_dir)
    new_index = ReportIndex.read(new_dir)
    assert new_index.entries["foo"].load().status == Status.PASS
    assert new_index.entries["bar"].load().status == Status.FAIL


def test_index_save_and_add(tmp_path):
    """Check that ReportIndex.save_and_add_result works as expected."""
    # Write a report with two results
    index = ReportIndex(str(tmp_path))
    index.save_and_add_result(Result("foo", status=Status.PASS))
    index.save_and_add_result(Result("bar", status=Status.FAIL))
    index.write()

    # Read it and make sure the results can be read
    index = ReportIndex.read(str(tmp_path))
    assert index.entries["foo"].load().status == Status.PASS
    assert index.entries["bar"].load().status == Status.FAIL


class TestMemleak:
    """Check the absence of memory leaks in a testsuite run."""

    class MySuite(MultiSchedulingSuite):
        def custom_adjust_dag_deps(self, dag):
            # Create our mapping to test fragments for convenience below
            self.test_fragments = {}
            for fg in dag.vertex_data.values():
                if fg.matches(TestMemleak.MyDriver, "run_wrapper"):
                    self.test_fragments[fg.driver.test_name] = fg
            t1 = self.test_fragments["t1"]
            t2 = self.test_fragments["t2"]

            # Create weakrefs to critical objects so that we can check later,
            # after a full GC collection, that no references to them are left.
            self.test_wr = weakref.WeakValueDictionary()

            # Make sure that t2 is executed after t1
            dag.update_vertex(vertex_id=t2.uid, predecessors=[t1.uid])

            # Flags to make sure that conditional code in collect_result were
            # executed.
            self.t1_result_collected = False
            self.test_succeeded = False

        def collect_result(self, fragment):
            # Before the actual "collect_result" method has a chance to remove
            # the last references to various data that we want to check, create
            # weak references to them.
            if fragment.uid is self.test_fragments["t1"].uid:
                driver = fragment.driver
                result = driver.result

                for key, value in [
                    ("t1_driver", driver),
                    ("t1_result_log", result.log),
                    # The "output" attribute is set during the test execution,
                    # so in multiprocessing mode, it is never set for the
                    # TestDriver instance in the testsuite "master" process.
                    ("t1_output", getattr(driver, "output", None)),
                    ("t1_result_out", result.out),
                    ("t1_result_expected", result.expected),
                ]:
                    if value is not None:
                        self.test_wr[key] = value
                self.t1_result_collected = True

            super().collect_result(fragment)

            # By the time we collect results for t2, all the refs for t1 should
            # be gone (they used to leak until the testsuite itself got
            # collected). Trigger a GC collection and check the absence of
            # weakref, which proves the absence of refs left to the critical
            # objects.
            if fragment.uid == self.test_fragments["t2"].uid:
                gc.collect()
                assert dict(self.test_wr) == {}
                self.test_succeeded = True

    class MyDriver(DiffDriver):
        @property
        def baseline(self):
            return (None, "", False)

        def run(self):
            self.result.log += "This is some log"
            self.output += "This is some output"

    def run_check(self, multiprocessing):
        suite = run_testsuite(
            create_testsuite(
                ["t1", "t2"],
                self.MyDriver,
                self.MySuite,
                self.MySuite.custom_adjust_dag_deps,
            ),
            ["--failure-exit-code=0"],
            multiprocessing=multiprocessing,
        )
        assert extract_results(suite) == {"t1": Status.FAIL, "t2": Status.FAIL}
        assert suite.t1_result_collected
        assert suite.test_succeeded

    def test_multithread(self):
        self.run_check(multiprocessing=False)

    def test_multiprocess(self):
        self.run_check(multiprocessing=True)


class TestSkipPassed:
    """Check that --skip-passed works as expected."""

    class MyDriver(BasicDriver):
        def run(self, prev, slot):
            pass

        def analyze(self, prev, slot):
            _, status_name = self.test_name.split("-")
            status = Status[status_name.upper()]
            self.result.set_status(status)
            self.push_result()

    def test(self):
        # Run the testsuite a first time. All tests should be scheduled and
        # thus be present in the report.
        tests_1 = [
            "t1-pass",
            "t1-fail",
            "t1-xfail",
            "t1-xpass",
            "t1-verify",
            "t1-skip",
            "t1-not_applicable",
            "t1-error",
        ]
        suite = run_testsuite(
            create_testsuite(tests_1, self.MyDriver),
            ["--skip-passed"],
            expect_failure=True,
        )
        assert extract_results(suite) == {
            "t1-pass": Status.PASS,
            "t1-fail": Status.FAIL,
            "t1-xfail": Status.XFAIL,
            "t1-xpass": Status.XPASS,
            "t1-verify": Status.VERIFY,
            "t1-skip": Status.SKIP,
            "t1-not_applicable": Status.NOT_APPLICABLE,
            "t1-error": Status.ERROR,
        }

        # Run the testsuite a second time. All tests but PASS, XFAIL and XPASS
        # ones should run, and the new one should run, too.
        tests_2 = tests_1 + ["t2-pass"]
        suite = run_testsuite(
            create_testsuite(tests_2, self.MyDriver),
            ["--skip-passed"],
            expect_failure=True,
        )
        assert extract_results(suite) == {
            "t1-fail": Status.FAIL,
            "t1-verify": Status.VERIFY,
            "t1-skip": Status.SKIP,
            "t1-not_applicable": Status.NOT_APPLICABLE,
            "t1-error": Status.ERROR,
            "t2-pass": Status.PASS,
        }


def test_filtering(tmp_path):
    """Check support for filtering tests on the command line."""
    # Create the hierarchy of test directories in the "tests" directory
    tests_dir = tmp_path / "tests"
    for tc, filename in [
        ("foo/a", "test.yaml"),
        ("foo/bar", "test.yaml"),
        ("bar", "test.yaml"),
        ("custom", "custom.yaml"),
    ]:
        tc_dir = tests_dir / tc
        mkdir(str(tc_dir))
        with (tc_dir / filename).open("w"):
            pass

    # This test finder, unlike YAMLTestFinder, does not guarantee that each
    # tests has its own directory.
    class CustomTestFinder(Finder):
        test_dedicated_directory = False

        def probe(self, testsuite, dirpath, dirnames, filenames):
            if "custom.yaml" in filenames:
                return [
                    ParsedTest(
                        f"custom.{t}",
                        MyDriver,
                        {},
                        dirpath,
                        test_matcher=f"custom.{t}",
                    )
                    for t in ["a", "b", "foo"]
                ]

    # Dummy test driver that always passes
    class MyDriver(BasicDriver):
        def run(self, prev, slot):
            pass

        def analyze(self, prev, slot):
            self.result.set_status(Status.PASS)
            self.push_result()

    class Mysuite(Suite):
        test_driver_map = {"mydriver": MyDriver}
        default_driver = "mydriver"
        tests_subdir = str(tests_dir)
        test_finders = [YAMLTestFinder(), CustomTestFinder()]

    def check(patterns, expected_results):
        """Run the testsuite with the given patterns.

        Check that we get exactly the expected set of test results.
        """
        suite = run_testsuite(Mysuite, patterns)
        assert extract_results(suite) == {
            name: Status.PASS for name in expected_results
        }

    # No pattern: all tests should run
    check(
        [], {"foo__a", "foo__bar", "bar", "custom.a", "custom.b", "custom.foo"}
    )

    # Pattern that is actually a directory: run only tests from that directory
    check([str(tests_dir / "bar")], {"bar"})

    # Pattern that matches a directory: run all tests that matches that pattern
    # anyway.
    check(["foo"], {"foo__a", "foo__bar", "custom.foo"})

    check(["custom"], {"custom.a", "custom.b", "custom.foo"})
    check([str(tests_dir / "custom")], {"custom.a", "custom.b", "custom.foo"})

    # Pattern that matches a "test matcher" string: run only the corresponding
    # test.
    check(["custom.b"], {"custom.b"})

    # Pattern that matches the middle of multiple tests: run only these tests
    check(["a"], {"foo__a", "foo__bar", "bar", "custom.a"})


def test_symlink_loop(tmp_path):
    """Check that testcase discovery completes even with symlink loops."""
    # Create the hierarchy of test directories in the "tests" directory, with
    # one subdirectory containing a symbolic link to the "tests" directory
    # itself.
    tests_dir = tmp_path / "tests"
    tests_dir.mkdir()
    for t in ["a", "b"]:
        tc_dir = tests_dir / t
        tc_dir.mkdir()
        with (tc_dir / "test.yaml").open("w"):
            pass

    (tests_dir / "a" / "link_loop").symlink_to(tests_dir)

    # Dummy test driver that always passes
    class MyDriver(BasicDriver):
        def run(self, prev, slot):
            pass

        def analyze(self, prev, slot):
            self.result.set_status(Status.PASS)
            self.push_result()

    class Mysuite(Suite):
        test_driver_map = {"mydriver": MyDriver}
        default_driver = "mydriver"
        tests_subdir = str(tests_dir)

    # Trying to run the testsuite used to create an infinite recursion in
    # testcase discovery code as the symbolic link was traversed again and
    # again.
    suite = run_testsuite(Mysuite, ["-E"])
    assert extract_results(suite) == {"a": Status.PASS, "b": Status.PASS}



"""Helpers to generate testsuite reports compatible with GAIA.

GAIA is AdaCore's internal Web analyzer.
"""

from __future__ import annotations

import dataclasses
import os.path
import tempfile
from typing import Dict, List, Optional, Set, TYPE_CHECKING, Union

import e3.env
from e3.testsuite.result import FailureReason, Log, TestResult, TestStatus

# Import TestsuiteCore only for typing, as this creates a circular import
if TYPE_CHECKING:
    from e3.testsuite.report.index import ReportIndex


STATUS_MAP = {
    TestStatus.PASS: "OK",
    TestStatus.FAIL: "FAIL",
    TestStatus.XFAIL: "XFAIL",
    TestStatus.XPASS: "UOK",
    TestStatus.VERIFY: "VERIFY",
    TestStatus.SKIP: "DEAD",
    TestStatus.NOT_APPLICABLE: "NOT-APPLICABLE",
    TestStatus.ERROR: "PROBLEM",
}
"""
Map TestStatus values to GAIA-compatible test statuses.
"""

FAILURE_REASON_MAP = {
    FailureReason.CRASH: "CRASH",
    FailureReason.TIMEOUT: "TIMEOUT",
    FailureReason.MEMCHECK: "PROBLEM",
    FailureReason.DIFF: "DIFF",
}
"""
Map FailureReason values to equivalent GAIA-compatible test statuses.
"""


def gaia_status(
    status: TestStatus, failure_reasons: Set[FailureReason]
) -> str:
    """Return the GAIA-compatible status that describes a result the best.

    :param status: Status for the result.
    :param failure_reasons: Set of failure reason for the result.
    """
    # Translate test failure status to the GAIA status that is most appropriate
    # given the failure reasons.
    if status == TestStatus.FAIL and failure_reasons:
        for reason in FailureReason:
            if reason in failure_reasons:
                return FAILURE_REASON_MAP[reason]
    return STATUS_MAP[status]


@dataclasses.dataclass(frozen=True)
class GAIAResultFiles:
    """Filenames for a given result in a GAIA report.

    In a GAIA testsuite report, each result is described as one entry (line) in
    the "results" text file, and several optional files. This object contains
    the names for the files, when present, corresponding to a given result.
    """

    result: Optional[str]
    log: Optional[str]
    expected: Optional[str]
    out: Optional[str]
    diff: Optional[str]
    time: Optional[str]
    info: Optional[str]


def dump_result_logs(result: TestResult, output_dir: str) -> GAIAResultFiles:
    """Write log files to describe a result in a GAIA report.

    :param result: Result to describe in the GAIA report.
    :param output_dir: Directory where to create the various files involved.
    """
    log_file: Optional[str] = None
    expected_file: Optional[str] = None
    out_file: Optional[str] = None
    diff_file: Optional[str] = None
    time_file: Optional[str] = None
    info_file: Optional[str] = None

    def unwrap_log(log: Union[Log, str, bytes]) -> Union[str, bytes]:
        return log.log if isinstance(log, Log) else log

    def write_log(log: Union[str, bytes], file_ext: str) -> str:
        mode: str
        encoding: Optional[str]

        mode, encoding = (
            ("wb", None) if isinstance(log, bytes) else ("w", "utf-8")
        )

        with tempfile.NamedTemporaryFile(
            prefix="gaia-result-",
            suffix=file_ext,
            dir=output_dir,
            mode=mode,
            encoding=encoding,
            delete=False,
        ) as f:
            f.write(log)
            return f.name

    status = gaia_status(result.status, result.failure_reasons)
    comment = result.msg or ""
    result_file = write_log(f"{status}:{comment}\n", ".result")

    if result.log:
        log = unwrap_log(result.log)
        assert isinstance(log, str)
        log_file = write_log(log, ".log")

    if result.expected is not None:
        expected = unwrap_log(result.expected)
        assert isinstance(expected, (str, bytes))
        expected_file = write_log(expected, ".expected")

    if result.out is not None:
        out = unwrap_log(result.out)
        assert isinstance(out, (str, bytes))
        out_file = write_log(out, ".out")

    if result.diff is not None:
        diff = unwrap_log(result.diff)
        assert isinstance(diff, (str, bytes))
        diff_file = write_log(diff, ".diff")

    if result.time is not None:
        # Nanoseconds granularity (9 decimals for seconds) should be
        # enough for any valuable time measurement. Rounding allows
        # predictable floating-point value representation.
        time_file = write_log("{:.9f}".format(result.time), ".time")

    if result.info:
        # Sort entries to have a deterministic output
        info_file = write_log(
            "\n".join(
                "{}:{}".format(key, value)
                for key, value in sorted(result.info.items())
            ),
            ".info",
        )

    return GAIAResultFiles(
        result_file,
        log_file,
        expected_file,
        out_file,
        diff_file,
        time_file,
        info_file,
    )


def dump_result_logs_if_needed(
    env: e3.env.Env,
    result: TestResult,
    output_dir: str,
) -> Optional[GAIAResultFiles]:
    """Shortcut to call dump_result_logs if a GAIA report is requested."""
    return (
        dump_result_logs(result, output_dir)
        if env.options.gaia_output is not None
        else None
    )


def dump_discriminants(discs: object, output_dir: str) -> None:
    """
    Dump discriminants for a GAIA-compatible testsuite report.

    :param discs: List of discriminants to dump.  Just like OptFileParse,
        accept either a string (comma-separated list of discriminant names) or
        a list of strings (list of discriminant names). Do nothing in other
        cases.
    :param output_dir: Directory in which to emit the report.
    """
    # Create the "discs" file if we had a supported type for "discs", even if
    # the list is empty, but don't create this file otherwise.
    discs_list: Optional[List[str]] = None
    if isinstance(discs, str):
        discs_list = discs.split(",")
    elif isinstance(discs, list) and all(isinstance(d, str) for d in discs):
        discs_list = discs
    if discs_list is not None:
        with open(
            os.path.join(output_dir, "discs"), "w", encoding="utf-8"
        ) as f:
            if discs_list:
                f.write(" ".join(discs_list))
                f.write("\n")


def dump_gaia_report(
    report_index: ReportIndex,
    output_dir: str,
    discs: object = None,
    result_files: Optional[Dict[str, GAIAResultFiles]] = None,
) -> None:
    """Dump a GAIA-compatible testsuite report.

    :param report_index: ReportIndex instance for all the test results to
        include in the report.
    :param output_dir: Directory in which to emit the report.
    :param discs: List of discriminants associated to the testsuite report, if
        any. See "dump_discriminants" for the expected format.
    :param result_files: If the log files for each result have already been
        generated, mapping from test names to result file names. None
        otherwise.
    """
    # If the result files are already generated, make sure we have exactly one
    # set of files per test result.
    if result_files is not None:
        assert set(result_files) == set(report_index.entries)

    # If there is a list of discriminants (i.e. in legacy AdaCore testsuites:
    # see AdaCoreLegacyTestDriver), include it in the report.
    dump_discriminants(discs, output_dir)

    with open(
        os.path.join(output_dir, "results"), "w", encoding="utf-8"
    ) as results_fd:
        for entry in report_index.entries.values():
            # Add an entry for it in the "results" index file
            message = entry.msg or ""
            status = gaia_status(entry.status, entry.failure_reasons)
            results_fd.write(f"{entry.test_name}:{status}:{message}\n")

            # Generate result files if they are not generated yet, then move
            # them where expected.
            files = (
                dump_result_logs(entry.load(), output_dir)
                if result_files is None
                else result_files[entry.test_name]
            )
            for ext, filename in dataclasses.asdict(files).items():
                if filename is not None:
                    new_filename = os.path.join(
                        output_dir, f"{entry.test_name}.{ext}"
                    )
                    os.rename(filename, new_filename)



"""Tests for the e3.testsuite.report.gaia module."""

import os

from e3.testsuite.report.gaia import dump_gaia_report
from e3.testsuite.result import FailureReason, TestStatus as Status

from .utils import create_report, create_result


def check(results, tmp_path, expected_files):
    """Generate a GAIA report and check its contents."""
    # Prepare output directories
    index_dir = tmp_path / "index"
    gaia_report_dir = tmp_path / "report"
    index_dir.mkdir()
    gaia_report_dir.mkdir()

    # Write test entries on disk
    index = create_report(results, str(index_dir))

    # Generate the GAIA report from them
    dump_gaia_report(index, str(gaia_report_dir))

    actual_files = {}
    for filename in os.listdir(gaia_report_dir):
        with open(gaia_report_dir / filename) as f:
            actual_files[filename] = f.read()

    assert actual_files == expected_files


def test_basics(tmp_path):
    """Test basic GAIA report features."""
    check(
        [
            create_result(
                "foo1",
                Status.PASS,
                msg="foo1 is ok",
                log="Logs for foo1.\n",
            ),
            create_result(
                "foo2",
                Status.FAIL,
                msg="baseline diff",
                out="foo\nbar\n",
                expected="foo\nbar\nbaz\n",
                diff="<some-diff>",
                failure_reasons={FailureReason.DIFF},
            ),
            create_result(
                "foo3",
                Status.XFAIL,
                msg="",
                failure_reasons={FailureReason.CRASH},
            ),
        ],
        tmp_path,
        {
            "results": (
                "foo1:OK:foo1 is ok\n"
                "foo2:DIFF:baseline diff\n"
                "foo3:XFAIL:\n"
            ),
            "foo1.result": "OK:foo1 is ok\n",
            "foo1.log": "Logs for foo1.\n",
            "foo2.result": "DIFF:baseline diff\n",
            "foo2.out": "foo\nbar\n",
            "foo2.expected": "foo\nbar\nbaz\n",
            "foo2.diff": "<some-diff>",
            "foo3.result": "XFAIL:\n",
        },
    )



"""Tests for the e3.testsuite.driver.classic module."""

import glob
import os
import os.path
import re
import sys

from e3.testsuite import Testsuite as Suite
import e3.testsuite.control as crtl
import e3.testsuite.driver.classic as classic
from e3.testsuite.report.index import ReportIndex
from e3.testsuite.result import TestStatus as Status

from .utils import (
    MultiSchedulingSuite,
    create_testsuite,
    extract_results,
    run_testsuite,
    suite_logs,
)


def test_control_interpret():
    """Check that TestControl.interpret works as expected."""
    # Mock data structures so that we don't have to create full testsuite runs

    class MockEnv:
        pass

    class MockDriver:
        def __init__(self, test_env, env):
            self.test_env = test_env
            self.env = MockEnv()
            for key, value in env.items():
                setattr(self.env, key, value)

        def test_dir(self, *args):
            return os.path.join("/nosuchdir", *args)

    def expect_result(test_env, condition_env=None, env=None):
        condition_env = condition_env if condition_env is not None else {}
        env = env if env is not None else {}
        driver = MockDriver(test_env, env)
        control = crtl.YAMLTestControlCreator(condition_env).create(driver)
        return control.skip, control.xfail, control.message

    def expect_error(test_env, condition_env=None, env=None):
        condition_env = condition_env if condition_env is not None else {}
        env = env if env is not None else {}
        driver = MockDriver(test_env, env)
        try:
            crtl.YAMLTestControlCreator(condition_env).create(driver)
        except ValueError as exc:
            return str(exc)
        else:
            raise AssertionError("exception expected")

    # No control entry: no test control
    assert expect_result({}) == (False, False, None)

    # Empty control list: no test control
    assert expect_result({"control": []}) == (False, False, None)

    # Invalid control object: error
    assert expect_error({"control": None}) == "list expected at the top level"

    # Invalid control entry: error
    assert (
        expect_error({"control": [[True]]})
        == "entry #1: list of 2 or 3 strings expected"
    )
    assert (
        expect_error({"control": [["SKIP", True]]})
        == "entry #1: list of 2 or 3 strings expected"
    )
    assert (
        expect_error({"control": [["no-such-control", "True"]]})
        == "entry #1: invalid kind: no-such-control"
    )
    assert expect_error({"control": [["SKIP", "foobar"]]}) == (
        "entry #1: invalid condition (NameError):"
        " name 'foobar' is not defined"
    )

    # Precedence to the first control whose condition is true
    assert expect_result(
        {
            "control": [
                ["SKIP", "False", "entry 1"],
                ["XFAIL", "True", "entry 2"],
                ["SKIP", "False", "entry 3"],
                ["SKIP", "True", "entry 4"],
            ]
        }
    ) == (False, True, "entry 2")

    # Use variables from test_env
    assert expect_result(
        {"control": [["SKIP", "foobar"]]}, condition_env={"foobar": True}
    ) == (True, False, None)
    assert expect_result(
        {"control": [["SKIP", "foobar"]]}, condition_env={"foobar": False}
    ) == (False, False, None)

    # Use variables from env
    assert expect_result(
        {"control": [["SKIP", "env.foobar"]]}, env={"foobar": True}
    ) == (True, False, None)
    assert expect_result(
        {"control": [["SKIP", "env.foobar"]]}, env={"foobar": False}
    ) == (False, False, None)


class ScriptDriver(classic.ClassicTestDriver):
    """Driver to run a Python script through ClassicTestDriver.shell."""

    helper_script = os.path.abspath(
        os.path.join(os.path.dirname(__file__), "classic-tests", "script.py")
    )

    @property
    def copy_test_directory(self):
        return self.test_env.get("copy_test_directory", True)

    def set_up(self):
        super(ScriptDriver, self).set_up()
        if not self.copy_test_directory:
            os.mkdir(self.working_dir())

        self.failures = []

        try:
            self.process_config = self.test_env["process"]
        except KeyError as exc:
            raise classic.TestAbortWithError(
                "Missing 'process' test.yaml entry"
            ) from exc

    def run(self):
        if self.test_env.get("force-skip", False):
            raise classic.TestSkip("forcing skip")

        config = self.process_config
        args = list(config["args"])
        if config.get("read"):
            args.append("-read={}".format(self.test_dir(config["read"])))
        p = self.shell(
            [sys.executable, self.helper_script] + args,
            catch_error=config.get("catch_error", True),
        )
        if p.out:
            self.failures.append("non-empty output")

    def compute_failures(self):
        return self.failures


class DummyDriver(classic.ClassicTestDriver):
    """Simple driver to check compute_failures's default behavior."""

    def run(self):
        pass


class TestClassic:
    """Check that ClassicTestDriver works as expected."""

    class Mysuite(MultiSchedulingSuite):
        tests_subdir = "classic-tests"
        test_driver_map = {
            "script-driver": ScriptDriver,
            "dummy-driver": DummyDriver,
        }

    def run_check(self, caplog, multiprocessing):
        # Look for tests using a relative path to check that test directories
        # are properly converted to absolute paths (see abs-test-dir).
        tests_subdir = os.path.relpath(
            os.path.join(
                os.path.abspath(os.path.dirname(__file__)),
                self.Mysuite.tests_subdir,
            ),
        )

        suite = run_testsuite(
            self.Mysuite,
            args=["--truncate-logs=0", tests_subdir, "-E"],
            multiprocessing=multiprocessing,
            expect_failure=True,
        )
        assert extract_results(suite) == {
            "simple": Status.PASS,
            "catch-error-pass": Status.PASS,
            "catch-error-fail": Status.FAIL,
            "skipped": Status.SKIP,
            "xfailed": Status.XFAIL,
            "xpassed": Status.XPASS,
            "errored": Status.ERROR,
            "force-skip": Status.SKIP,
            "invalid-control": Status.ERROR,
            "dummy": Status.PASS,
            "with-output": Status.FAIL,
            "binary-output": Status.FAIL,
            "invalid-utf8-output": Status.ERROR,
            "suspicious-test-opt": Status.PASS,
            "abs-test-dir": Status.PASS,
            "long-logs": Status.FAIL,
        }

        log = (
            'suspicious-test-opt: "test.opt" file found whereas only "control"'
            " entries are considered"
        )
        assert log in suite_logs(caplog)

        # Check that XPASS result prints the XFAIL message
        assert suite.report_index.entries["xpassed"].msg == "XFAIL message"

    def test_multithreading(self, caplog):
        self.run_check(caplog, multiprocessing=False)

    # TODO: somehow forward logs emitted in multiprocess test fragments to the
    # testsuite's main logging system.


def test_long_logs(caplog):
    """Check that long logs are truncated as requested."""

    class Mysuite(Suite):
        tests_subdir = "classic-tests"
        test_driver_map = {"script-driver": ScriptDriver}

    suite = run_testsuite(
        Mysuite,
        args=["--truncate-logs=3", "long-logs", "--gaia-output"],
        expect_failure=True,
    )
    assert extract_results(suite) == {"long-logs": Status.FAIL}

    with open(os.path.join("out", "new", "long-logs.log")) as f:
        content = f.read().splitlines()
    assert content[0].startswith("Running: ")
    assert content[1:] == [
        "Status code: 0",
        "Output: ",
        "a0",
        "a1",
        "a2",
        "",
        "... 8 lines skipped...",
        "",
        "b4",
        "b5",
        "b6",
    ]


class TestCleanupMode:
    """Check that --cleanup-mode works as expected."""

    class ControlCreator(crtl.TestControlCreator):
        def create(self, driver):
            name = driver.test_name
            return crtl.TestControl(xfail="xpass" in name or "xfail" in name)

    class MyDriver(classic.ClassicTestDriver):
        copy_test_directory = False

        @property
        def test_control_creator(self):
            return TestCleanupMode.ControlCreator()

        def run(self):
            os.mkdir(self.working_dir())
            with open(self.working_dir("flag.txt"), "w"):
                pass

            if "error" in self.test_name:
                raise classic.TestAbortWithError("error")
            if "fail" in self.test_name:
                raise classic.TestAbortWithFailure("failure")

    EXPECTED_RESULTS = {
        "pass": Status.PASS,
        "with_fail": Status.FAIL,
        "with_xfail": Status.XFAIL,
        "with_xpass": Status.XPASS,
        "with_error": Status.ERROR,
    }

    FAILING_TESTS = {"with_fail", "with_xfail", "with_xpass", "with_error"}
    ALL_TESTS = set(EXPECTED_RESULTS)

    def run(self, tmp_path, args, expected_dirs):
        suite = run_testsuite(
            create_testsuite(list(self.EXPECTED_RESULTS), self.MyDriver),
            args=args + ["-t", str(tmp_path)],
            expect_failure=True,
        )
        assert extract_results(suite) == self.EXPECTED_RESULTS

        # Compute the list of working spaces left, i.e. directories that
        # contain the "flag.txt" file.
        working_dirs = {
            os.path.basename(os.path.dirname(d))
            for d in glob.glob(str(tmp_path / "*" / "*" / "flag.txt"))
        }

        assert working_dirs == expected_dirs

    # Check the default behavior
    def test_default(self, tmp_path):
        self.run(tmp_path, [], self.FAILING_TESTS)

    # Check all possible --cleanup-mode values
    def test_cm_none(self, tmp_path):
        self.run(tmp_path, ["--cleanup-mode=none"], self.ALL_TESTS)

    def test_cm_passing(self, tmp_path):
        self.run(tmp_path, ["--cleanup-mode=passing"], self.FAILING_TESTS)

    def test_cm_all(self, tmp_path):
        self.run(tmp_path, ["--cleanup-mode=all"], set())

    # Check --disable-cleanup alone, and that --cleanup-mode has precedence
    def test_dc(self, tmp_path):
        self.run(tmp_path, ["--disable-cleanup"], self.ALL_TESTS)

    def test_cm_dc(self, tmp_path):
        self.run(
            tmp_path,
            ["--disable-cleanup", "--cleanup-mode=all"],
            set(),
        )


class TestCleanupFailure:
    """Check that error recovery for working dir cleanup works as expected."""

    class MyDriver(classic.ClassicTestDriver):
        def run(self):
            os.mkdir(self.working_dir("foo"))
            with open(self.working_dir("foo", "bar.txt"), "w"):
                pass

        def cleanup_working_dir(self):
            raise RuntimeError("some cleanup failure")

    class Mysuite(Suite):
        tests_subdir = "simple-tests"

        @property
        def test_driver_map(self):
            return {"default": TestCleanupFailure.MyDriver}

        default_driver = "default"

    def test(self):
        suite = run_testsuite(self.Mysuite, expect_failure=True)
        assert extract_results(suite) == {
            "test1": Status.PASS,
            "test2": Status.PASS,
            "test1__tear_down": Status.ERROR,
            "test2__tear_down": Status.ERROR,
        }

        index = ReportIndex.read("out/new")
        r = index.entries["test1__tear_down"].load()

        assert re.match(
            "Error while removing the working directory .*test1:\n"
            "\n"
            "Traceback (?:.|\n)*\n"
            "RuntimeError: some cleanup failure\n"
            "\n"
            "Remaining files:\n"
            f"  test.yaml\n"
            f"  foo\n"
            f"  foo{os.path.sep}bar.txt\n",
            r.log,
        )


class TestIgnoreEnv:
    """Check that shell's ignore_env argument works as expected."""

    class MyDriver(classic.ClassicTestDriver):
        def run(self):
            self.shell(
                [sys.executable, self.env.env_printer],
                env={"E3_TESTSUITE_VAR2": "var2-value"},
                **self.env.kwargs,
            )
            self.result.out = self.output

    def run(self, expected_output, **kwargs):
        env_printer = self.env_printer

        class Mysuite(Suite):
            tests_subdir = "simple-tests"
            test_driver_map = {"default": TestIgnoreEnv.MyDriver}
            default_driver = "default"

            def set_up(self):
                self.env.env_printer = env_printer
                self.env.kwargs = kwargs

        suite = run_testsuite(Mysuite, args=["-E", "test1"])
        assert extract_results(suite) == {"test1": Status.PASS}

        index = ReportIndex.read("out/new")
        r = index.entries["test1"].load()
        lines = [line.strip() for line in r.out.splitlines()]
        assert lines == expected_output

    def test(self):
        self.env_printer = os.path.join(
            os.path.dirname(os.path.abspath(__file__)), "env_printer.py"
        )
        os.environ["E3_TESTSUITE_VAR1"] = "var1-value"

        self.run(
            ["E3_TESTSUITE_VAR1=var1-value", "E3_TESTSUITE_VAR2=var2-value"],
            ignore_environ=False,
        )
        self.run(["E3_TESTSUITE_VAR2=var2-value"], ignore_environ=True)
        self.run(["E3_TESTSUITE_VAR2=var2-value"])


class TestMayHaveTimedOut:
    """Check the ClassicTestDriver.process_may_have_timed_out method."""

    class MyDriver(classic.ClassicTestDriver):
        def run(self):
            def check(status, output):
                return self.process_may_have_timed_out(
                    classic.ProcessResult(status, output)
                )

            matching_msg = "rlimit: Real time limit (1 s) exceeded\n"

            assert not check(0, "foobar")
            assert not check(2, "foobar")
            assert not check(0, matching_msg)
            assert check(2, matching_msg)

    def test(self):
        class Mysuite(Suite):
            tests_subdir = "simple-tests"
            test_driver_map = {"default": self.MyDriver}
            default_driver = "default"

        suite = run_testsuite(Mysuite, args=["-E", "test1"])
        assert extract_results(suite) == {"test1": Status.PASS}


def test_decoding_error(caplog):
    """Check that process output decoding errors are properly reported."""

    class MyDriver(ScriptDriver):
        def set_up(self):
            self.test_env["process"] = {"args": ["-b"]}
            super().set_up()

    suite = run_testsuite(
        create_testsuite(["t"], MyDriver),
        args=["-E"],
        expect_failure=True,
    )
    assert extract_results(suite) == {"t": Status.ERROR}
    log = suite.report_index.entries["t"].load().log
    assert re.match(
        "Running: .*script.py -b \\(cwd=.*\\)"
        "\nCannot decode subprocess output:"
        "\n"
        "\n  h\\\\xe9llo",
        log,
    )


def test_shell_stdin():
    """Check passing non-default stdin paramater to ClassicTestDriver.shell."""

    class MyDriver(classic.ClassicTestDriver):
        def run(self):
            self.shell(
                [sys.executable, ScriptDriver.helper_script, "-stdin"],
                stdin=os.path.join(
                    os.path.dirname(__file__),
                    "classic-tests",
                    "abs-test-dir",
                    "input.txt",
                ),
            )
            self.result.out = self.output

    suite = run_testsuite(create_testsuite(["t"], MyDriver), args=["-E"])
    assert extract_results(suite) == {"t": Status.PASS}
    r = suite.report_index.entries["t"].load()
    assert r.out == "From stdin: 'This is the content of input.txt\\n'\n"



"""Tests for the "status" file."""

import os

from e3.testsuite import Testsuite as Suite
from e3.testsuite.driver import TestDriver as Driver
from e3.testsuite.result import TestResult as Result, TestStatus as Status

from .utils import run_testsuite


class TestBasic:
    """Check that the test status at each step with no parallelism."""

    fragments = [
        ("frag_a", Status.PASS),
        ("frag_b", Status.FAIL),
        ("frag_c", Status.XFAIL),
        ("frag_d", Status.PASS),
    ]

    expected_statuses = {
        "frag_a": (
            "Test fragments: 0 / 4 completed"
            "\nCurrently running:"
            "\n  test1.frag_a"
            "\nPartial results:"
            "\n  <none>"
        ),
        "frag_b": (
            "Test fragments: 1 / 4 completed"
            "\nCurrently running:"
            "\n  test1.frag_b"
            "\nPartial results:"
            "\n  PASS         1"
        ),
        "frag_c": (
            "Test fragments: 2 / 4 completed"
            "\nCurrently running:"
            "\n  test1.frag_c"
            "\nPartial results:"
            "\n  PASS         1"
            "\n  FAIL         1"
        ),
        "frag_d": (
            "Test fragments: 3 / 4 completed"
            "\nCurrently running:"
            "\n  test1.frag_d"
            "\nPartial results:"
            "\n  PASS         1"
            "\n  FAIL         1"
            "\n  XFAIL        1"
        ),
        "final": (
            "Test fragments: 4 / 4 completed"
            "\nCurrently running:"
            "\n  <none>"
            "\nPartial results:"
            "\n  PASS         2"
            "\n  FAIL         1"
            "\n  XFAIL        1"
        ),
    }

    @staticmethod
    def check_status(label):
        with open(os.path.join("out", "new", "status")) as f:
            actual = f.read()
        assert (
            actual == TestBasic.expected_statuses[label]
        ), f"Unexpected status file at {label}: {actual}"

    class MyDriver(Driver):
        def add_test(self, dag):
            # Only for "test1", create a chain of fragments that just push some
            # result.
            if self.test_name == "test1":
                prev = None
                for name, _ in TestBasic.fragments:
                    self.add_fragment(
                        dag,
                        name,
                        None,
                        after=[prev] if prev else [],
                    )
                    prev = name

        def frag_a(self, prev, slot):
            self.run("frag_a", prev, slot)

        def frag_b(self, prev, slot):
            self.run("frag_b", prev, slot)

        def frag_c(self, prev, slot):
            self.run("frag_c", prev, slot)

        def frag_d(self, prev, slot):
            self.run("frag_d", prev, slot)

        def run(self, name, prev, slot):
            TestBasic.check_status(name)

            result = Result(name, self.test_env)
            for fragment_name, status in TestBasic.fragments:
                if fragment_name == name:
                    result.set_status(status)
                    break
            else:
                raise AssertionError(f"invalid fragment name: {name}")
            self.push_result(result)

    def test(self):
        class Mysuite(Suite):
            tests_subdir = "simple-tests"
            test_driver_map = {"default": self.MyDriver}
            default_driver = "default"

        run_testsuite(
            Mysuite,
            args=["-j1", "--show-error-output", "--status-update-interval=0"],
            expect_failure=True,
        )
        self.check_status("final")



import argparse
from typing import List, Optional, Set

from e3.testsuite.report.index import ReportIndex
from e3.testsuite.result import TestStatus


args_parser = argparse.ArgumentParser(
    description="Look for testcases that are always skipped in a set of"
    " testsuite reports."
)
args_parser.add_argument(
    "reports",
    metavar="RESULT_DIR",
    nargs="+",
    help="Directory that contains the report to load.",
)


def always_skipped_tests(reports: List[ReportIndex]) -> Set[str]:
    """Return the set of test names in ``reports`` that are always skipped."""
    skipped_once = set()
    not_skipped_once = set()
    for r in reports:
        for e in r.entries.values():
            if e.status == TestStatus.SKIP:
                skipped_once.add(e.test_name)
            else:
                not_skipped_once.add(e.test_name)
    return skipped_once - not_skipped_once


def main(argv: Optional[List[str]] = None) -> None:
    args = args_parser.parse_args(argv)
    tests = always_skipped_tests([ReportIndex.read(d) for d in args.reports])
    if tests:
        print("The following tests are never executed:")
        for t in sorted(tests):
            print("  {}".format(t))
    else:
        print("All testcases are executed at least once.")


if __name__ == "__main__":  # interactive-only
    main()



"""Tests for the "test.opt" handling."""

import os
import shutil
import tempfile

from e3.testsuite import Testsuite as Suite
import e3.testsuite.control as control
from e3.testsuite.driver.adacore import AdaCoreLegacyTestDriver as ACDriver
import e3.testsuite.testcase_finder as testcase_finder
from e3.testsuite.result import TestStatus as Status

from .utils import (
    MultiSchedulingSuite,
    extract_results,
    run_testsuite,
    suite_logs,
)


def test_adacore():
    """Check that AdaCore legacy support works as expected."""

    class Mysuite1(Suite):
        tests_subdir = "adacore-tests"

        def set_up(self):
            super(Mysuite1, self).set_up()
            self.env.discs = ["foo"]
            self.env.test_environ = dict(os.environ)

        test_finders = [testcase_finder.AdaCoreLegacyTestFinder(ACDriver)]

    suite = run_testsuite(Mysuite1, ["-E"], expect_failure=True)
    assert extract_results(suite) == {
        # Check that test.sh are picked up
        "0000-097": Status.PASS,
        # Check that test.sh aren't converted from cmd to Bourne syntax
        "3807-001": Status.PASS,
        # Check that test.cmd_suffix tests are converted
        "DB15-019": Status.PASS,
        # Regular test execution, exercize output refiners
        "T415-993": Status.PASS,
        # Missing non-default baseline
        "T415-994": Status.ERROR,
        # Missing test script
        "T415-995": Status.ERROR,
        # test.opt's SKIP yields a XFAIL without trying to execute the test
        "T415-996": Status.XFAIL,
        # test.opt says DEAD when there is a "foo" discriminant
        "T415-997": Status.SKIP,
        # Regular test execution and failing diff comparison (test.out absent)
        "T415-998": Status.FAIL,
        # Regular test execution and successful diff comparison
        "T415-999": Status.PASS,
        # Regular test execution and successful diff comparison
        "TA23-999": Status.PASS,
        # Regular test execution and failing diff comparison (test.out present)
        "Z999-999": Status.XFAIL,
    }

    class Mysuite2(Suite):
        tests_subdir = "adacore-support-tests"

        def set_up(self):
            super(Mysuite2, self).set_up()
            self.env.discs = ["FORCE_SH"]
            os.environ["TEST_SUPPORT_DIR"] = os.path.join(
                os.path.abspath(os.path.dirname(__file__)),
                "adacore-support-tests",
                "support",
            )
            self.env.test_environ = dict(os.environ)

        def tear_down(self):
            os.environ["TEST_SUPPORT_DIR"] = ""
            super(Mysuite2, self).tear_down()

        test_finders = [testcase_finder.AdaCoreLegacyTestFinder(ACDriver)]

    suite = run_testsuite(Mysuite2)
    assert extract_results(suite) == {"T415-999": Status.PASS}


def test_optfile(caplog):
    """Check that OptfileTestControlCreator works as expected."""

    class Mydriver(ACDriver):
        test_control_creator = control.AdaCoreLegacyTestControlCreator(["foo"])

    class Mysuite(MultiSchedulingSuite):
        tests_subdir = "adacore-tests"
        test_driver_map = {"adacore": Mydriver}
        default_driver = "adacore"

        def set_up(self):
            super().set_up()
            self.env.discs = []
            self.env.test_environ = dict(os.environ)

    suite = run_testsuite(Mysuite)
    assert extract_results(suite) == {
        "just-cmd": Status.PASS,
        "just-py": Status.PASS,
        "both-cmd-py": Status.PASS,
        "extra-control": Status.PASS,
    }

    message = (
        'extra-control: "control" entry found in test.yaml whereas only'
        " test.opt files are considered"
    )
    assert message in suite_logs(caplog)


def test_rewriting(caplog):
    """Check that ACDriver's baseline rewriting works as expected."""
    # This testcase involves the rewriting of testcase files, so work on a
    # temporary copy.
    with tempfile.TemporaryDirectory(
        prefix="test_adacore_rewriting"
    ) as temp_dir:
        tests_source = os.path.join(
            os.path.dirname(__file__), "adacore-rewriting-tests"
        )
        tests_copy = os.path.join(temp_dir, "tests")
        shutil.copytree(tests_source, tests_copy)

        class Mysuite(Suite):
            tests_subdir = tests_copy
            test_driver_map = {"adacore": ACDriver}
            default_driver = "adacore"

            def add_options(self, parser):
                parser.add_argument("--rewrite", "-r", action="store_true")

            def set_up(self):
                super().set_up()
                self.env.discs = []
                self.env.test_environ = dict(os.environ)
                self.env.rewrite_baselines = self.main.args.rewrite

        def check_baselines(test, expected_desc):
            files = sorted(
                f
                for f in os.listdir(os.path.join(tests_copy, test))
                if f.endswith(".out")
            )
            expected_files = sorted(expected_desc)
            assert files == expected_files

            for filename in files:
                expected_lines = expected_desc[filename]
                with open(os.path.join(tests_copy, test, filename)) as f:
                    lines = [line.rstrip() for line in f]
                assert lines == expected_lines

        # Make sure we have the expected baselines before running the testsuite
        check_baselines("default-nodiff", {"test.out": ["Hello"]})
        check_baselines("default-diff", {"test.out": ["World"]})
        check_baselines("default-empty", {"test.out": ["Hello"]})
        check_baselines("nondefault-nodiff", {"baseline.out": ["Hello"]})
        check_baselines("nondefault-empty", {"baseline.out": ["Hello"]})
        check_baselines("xfail-diff", {"test.out": ["World"]})

        suite = run_testsuite(Mysuite, args=["-r"], expect_failure=True)
        assert extract_results(suite) == {
            "default-nodiff": Status.PASS,
            "default-diff": Status.FAIL,
            "default-empty": Status.FAIL,
            "nondefault-nodiff": Status.PASS,
            "nondefault-empty": Status.FAIL,
            "xfail-diff": Status.XFAIL,
        }

        # Now check that baselines were updated as expected
        check_baselines("default-nodiff", {"test.out": ["Hello"]})
        check_baselines("default-diff", {"test.out": ["Hello"]})
        check_baselines("default-empty", {})
        check_baselines("nondefault-nodiff", {"baseline.out": ["Hello"]})
        check_baselines("nondefault-empty", {"baseline.out": []})
        check_baselines("xfail-diff", {"test.out": ["World"]})


def test_gaia_discs():
    """Check that the GAIA report reports discriminants as expected."""
    discs_filename = os.path.join("out", "new", "discs")

    class BasicSuite(Suite):
        tests_subdir = "simple-tests"

    def run(cls):
        run_testsuite(cls, args=["nosuchtest", "--gaia-output"])

    def test_discs(discs_value, report_discs):
        """Run a testsuite and check the list of discriminants in report.

        Run a testsuite with ``testsuite.env.discs = discs_value``. Check that
        there is no "discs" report file if ``report_discs`` is None, and that
        it is a text file that contains the ``report_discs`` string otherwise.
        """

        class DiscsSuite(Suite):
            def set_up(self):
                super().set_up()
                self.env.discs = discs_value

        run(DiscsSuite)
        if report_discs is None:
            assert not os.path.exists(discs_filename)
        else:
            assert os.path.exists(discs_filename)
            with open(discs_filename) as f:
                assert f.read() == report_discs

    # Check that no "discs" file is created when there is no "env.discs"
    # attribute, or when it does not contain a list of strings.
    run(BasicSuite)
    assert not os.path.exists(discs_filename)

    test_discs(None, None)
    test_discs(1, None)
    test_discs([1], None)

    # Check that this file is created with the expected content when
    # "env.discs" contains a list of strings.
    test_discs([], "")
    test_discs(["foo", "bar"], "foo bar\n")
    test_discs("a,b,c", "a b c\n")


def test_timeout():
    """Check that AdaCore legacy support detects timeouts."""

    class Mysuite(Suite):
        tests_subdir = "adacore-timeout-tests"
        test_driver_map = {"adacore": ACDriver}
        default_driver = "adacore"

        def set_up(self):
            super(Mysuite, self).set_up()
            self.env.discs = []
            self.env.test_environ = dict(os.environ)

    suite = run_testsuite(
        Mysuite, ["--gaia-output", "-E"], expect_failure=True
    )
    assert extract_results(suite) == {
        "pass": Status.PASS,
        "timedout": Status.FAIL,
    }

    with open(os.path.join("out", "new", "results")) as f:
        lines = sorted(f.read().splitlines())
        assert lines == [
            "pass:OK:",
            "timedout:TIMEOUT:unexpected output | test timed out",
        ]


def test_script_encoding():
    """Check that ACDriver correctly manages script encodings."""

    class MyDriver(ACDriver):
        @property
        def default_encoding(self):
            return self.test_env.get("encoding", "binary")

        @property
        def output_refiners(self):
            return []

    class Mysuite(Suite):
        tests_subdir = "adacore-script-encoding-tests"
        test_driver_map = {"mydriver": MyDriver}
        default_driver = "mydriver"

        def set_up(self):
            super(Mysuite, self).set_up()
            self.env.discs = []
            self.env.test_environ = dict(os.environ)

    suite = run_testsuite(Mysuite, ["-E"], expect_failure=True)
    results = extract_results(suite)
    assert results.pop("default") == Status.PASS
    assert results.pop("latin-1") == Status.PASS
    assert results.pop("utf-8") == Status.PASS
    assert len(results) == 1
    test_name, status = list(results.items())[0]
    assert test_name == "bad.run_wrapper"
    assert status == Status.ERROR
